# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration tests for remote replication modes."""

from __future__ import (
    annotations,
)
import os
from typing import (
    final,
)

import bzfs_main.detect
from bzfs_main.util.utils import (
    DIE_STATUS,
)
from bzfs_tests.itest import (
    ibase,
)
from bzfs_tests.itest.ibase import (
    SSH_PROGRAM,
    IntegrationTestCase,
    is_pv_at_least_1_9_0,
)
from bzfs_tests.itest.test_local import (
    LocalTestCase,
)
from bzfs_tests.tools import (
    stop_on_failure_subtest,
)
from bzfs_tests.zfs_util import (
    dataset_exists,
)


#############################################################################
class MinimalRemoteTestCase(IntegrationTestCase):

    def test_aaa_log_diagnostics_first(self) -> None:
        LocalTestCase(param=self.param).test_aaa_log_diagnostics_first()

    def test_basic_replication_flat_simple(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_simple()

    def test_basic_replication_recursive1(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_recursive1()

    def test_delete_dst_datasets_recursive_with_dummy_src(self) -> None:
        LocalTestCase(param=self.param).test_delete_dst_datasets_recursive_with_dummy_src()

    def test_zfs_set_via_recv_o(self) -> None:
        LocalTestCase(param=self.param).test_zfs_set_via_recv_o()

    def test_basic_replication_recursive_parallel(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_recursive_parallel()

    def test_inject_unavailable_sudo(self) -> None:
        expected_error = DIE_STATUS if os.getuid() != 0 and not self.is_no_privilege_elevation() else 0
        self.inject_unavailable_program("inject_unavailable_sudo", expected_error=expected_error)
        self.tearDownAndSetup()
        expected_error = 1 if os.getuid() != 0 and not self.is_no_privilege_elevation() else 0
        self.inject_unavailable_program("inject_failing_sudo", expected_error=expected_error)

    def test_disabled_sudo(self) -> None:
        expected_status = 0
        if os.getuid() != 0 and not self.is_no_privilege_elevation():
            expected_status = DIE_STATUS
        self.inject_disabled_program("sudo", expected_error=expected_status)

    def inject_disabled_program(self, prog: str, expected_error: int = 0) -> None:
        self.setup_basic()
        self.run_bzfs(
            ibase.SRC_ROOT_DATASET,
            ibase.DST_ROOT_DATASET,
            f"--{prog}-program=" + bzfs_main.detect.DISABLE_PRG,
            expected_status=expected_error,
        )
        if expected_error != 0:
            self.assert_snapshots(ibase.DST_ROOT_DATASET, 0)

    def inject_unavailable_program(self, *flags: str, expected_error: int = 0) -> None:
        self.setup_basic()
        inject_params: dict[str, bool] = {}
        for flag in flags:
            inject_params[flag] = True
        self.run_bzfs(
            ibase.SRC_ROOT_DATASET,
            ibase.DST_ROOT_DATASET,
            expected_status=expected_error,
            inject_params=inject_params,
        )
        if expected_error != 0:
            self.assert_snapshots(ibase.DST_ROOT_DATASET, 0)
        else:
            self.assert_snapshots(ibase.DST_ROOT_DATASET, 3, "s")


#############################################################################
@final
class FullRemoteTestCase(MinimalRemoteTestCase):

    def test_ssh_program_must_not_be_disabled_in_nonlocal_mode(self) -> None:
        LocalTestCase(param=self.param).test_ssh_program_must_not_be_disabled_in_nonlocal_mode()

    def test_basic_replication_flat_nothing_todo(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_nothing_todo()

    def test_basic_replication_without_source(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_without_source()

    def test_complex_replication_flat_use_bookmarks(self) -> None:
        LocalTestCase(param=self.param).test_complex_replication_flat_use_bookmarks()

    def test_basic_replication_flat_send_recv_flags(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_send_recv_flags()

    def test_basic_replication_flat_simple_with_multiple_root_datasets(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_simple_with_multiple_root_datasets()

    def test_basic_replication_dataset_with_spaces(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_dataset_with_spaces()

    def test_basic_replication_flat_with_multiple_root_datasets_converted_from_recursive(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_with_multiple_root_datasets_converted_from_recursive()

    def test_basic_replication_flat_with_ssh_exit_on_shutdown(self) -> None:
        LocalTestCase(param=self.param).test_basic_replication_flat_with_ssh_exit_on_shutdown()

    def test_zfs_set(self) -> None:
        LocalTestCase(param=self.param).test_zfs_set()

    def test_zfs_set_via_set_include(self) -> None:
        LocalTestCase(param=self.param).test_zfs_set_via_set_include()

    def test_inject_src_pipe_fail(self) -> None:
        self.inject_pipe_error("inject_src_pipe_fail", expected_error=[1, DIE_STATUS])

    def test_inject_src_pipe_garble(self) -> None:
        if is_pv_at_least_1_9_0():
            self.skipTest("workaround for zfs send-receive pipeline hang")
        self.inject_pipe_error("inject_src_pipe_garble")

    def test_inject_dst_pipe_garble(self) -> None:
        self.inject_pipe_error("inject_dst_pipe_garble")

    def test_inject_src_send_error(self) -> None:
        self.inject_pipe_error("inject_src_send_error")

    def test_inject_dst_receive_error(self) -> None:
        self.inject_pipe_error("inject_dst_receive_error", expected_error=2)

    def inject_pipe_error(self, flag: str, expected_error: int | list[int] = 1) -> None:
        self.setup_basic()
        for i in range(2):
            with stop_on_failure_subtest(i=i):
                inject_params: dict[str, bool] = {}
                if i == 0:
                    inject_params[flag] = True
                self.run_bzfs(
                    ibase.SRC_ROOT_DATASET,
                    ibase.DST_ROOT_DATASET,
                    expected_status=expected_error if i == 0 else 0,
                    inject_params=inject_params,
                )
                if i == 0:
                    self.assert_snapshots(ibase.DST_ROOT_DATASET, 0)
                else:
                    self.assert_snapshots(ibase.DST_ROOT_DATASET, 3, "s")

    def test_inject_unavailable_mbuffer(self) -> None:
        self.inject_unavailable_program("inject_unavailable_mbuffer")
        if self.param and self.param.get("ssh_mode") != "local" and self.param.get("min_pipe_transfer_size", -1) == 0:
            self.tearDownAndSetup()
            self.inject_unavailable_program("inject_failing_mbuffer", expected_error=1)

    def test_inject_unavailable_ps(self) -> None:
        self.inject_unavailable_program("inject_unavailable_ps")

    def test_inject_unavailable_pv(self) -> None:
        self.inject_unavailable_program("inject_unavailable_pv")

    def test_inject_unavailable_sh(self) -> None:
        self.inject_unavailable_program("inject_unavailable_sh")
        self.tearDownAndSetup()
        self.inject_unavailable_program("inject_failing_sh")

    def test_inject_unavailable_zstd(self) -> None:
        self.inject_unavailable_program("inject_unavailable_zstd")

    def test_inject_unavailable_zpool(self) -> None:
        self.inject_unavailable_program("inject_unavailable_zpool")
        self.tearDownAndSetup()
        self.inject_unavailable_program("inject_failing_zpool")

    def test_inject_unavailable_ssh(self) -> None:
        if self.param and self.param.get("ssh_mode") != "local":
            self.inject_unavailable_program("inject_unavailable_" + SSH_PROGRAM, expected_error=DIE_STATUS)
            self.tearDownAndSetup()
            self.inject_unavailable_program("inject_failing_" + SSH_PROGRAM, expected_error=DIE_STATUS)

    def test_inject_unavailable_zfs(self) -> None:
        self.inject_unavailable_program("inject_unavailable_zfs", expected_error=DIE_STATUS)
        self.tearDownAndSetup()
        self.inject_unavailable_program("inject_failing_zfs", expected_error=DIE_STATUS)

    def test_disabled_mbuffer(self) -> None:
        self.inject_disabled_program("mbuffer")

    def test_disabled_ps(self) -> None:
        self.inject_disabled_program("ps")

    def test_disabled_pv(self) -> None:
        self.inject_disabled_program("pv")

    def test_disabled_sh(self) -> None:
        self.inject_disabled_program("shell")

    def test_disabled_compression(self) -> None:
        self.inject_disabled_program("compression")

    def test_disabled_zpool(self) -> None:
        self.inject_disabled_program("zpool")

    def test_ssh_master_check_keeps_tcp_connection_alive_with_replication_recursive(self) -> None:
        self.setup_basic()
        self.run_bzfs(ibase.SRC_ROOT_DATASET, ibase.DST_ROOT_DATASET, "--recursive", ssh_control_persist_margin_secs=2**64)
        self.assert_snapshots(ibase.DST_ROOT_DATASET, 3, "s")
        self.assert_snapshots(ibase.DST_ROOT_DATASET + "/foo", 3, "t")
        self.assert_snapshots(ibase.DST_ROOT_DATASET + "/foo/a", 3, "u")
        self.assertFalse(dataset_exists(ibase.DST_ROOT_DATASET + "/foo/b"))  # b/c src has no snapshots



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for call_with_retries_async()."""

from __future__ import (
    annotations,
)
import asyncio
import functools
import logging
import pickle
import unittest
from collections.abc import (
    Awaitable,
)
from contextlib import (
    suppress,
)
from typing import (
    Callable,
    TypeVar,
)
from unittest.mock import (
    AsyncMock,
    MagicMock,
    patch,
)

from bzfs_main.util.retry import (
    NO_LOGGER,
    AttemptOutcome,
    Retry,
    RetryableError,
    RetryConfig,
    RetryError,
    RetryPolicy,
    RetryTemplate,
    RetryTiming,
    before_attempt_noop,
    call_with_retries_async,
)


#############################################################################
def suite() -> unittest.TestSuite:
    test_cases = [
        TestCallWithRetriesAsync,
        TestRetryTemplateWrapsAsync,
    ]
    return unittest.TestSuite(unittest.TestLoader().loadTestsFromTestCase(test_case) for test_case in test_cases)


#############################################################################
_T = TypeVar("_T")


async def asyncio_await_with_timeout(awaitable: Awaitable[_T], timeout_nanos: int, *, display_msg: object = "timeout") -> _T:
    """Convenience function that awaits an awaitable with a hard timeout; on timeout raises RetryableError.

    Assumes awaitable handles cancellation (CancelledError) correctly.
    """
    import asyncio

    if timeout_nanos < 0:
        raise ValueError(f"Invalid timeout_nanos: must be >= 0 but got {timeout_nanos}")
    if timeout_nanos == 0:
        raise RetryableError(display_msg=display_msg) from TimeoutError("Async operation timed out")

    task: asyncio.Future[_T] = asyncio.ensure_future(awaitable)
    try:
        done, pending = await asyncio.wait({task}, timeout=timeout_nanos / 1_000_000_000)
        if task in done:
            return task.result()
        assert task in pending
        task.cancel()
        with suppress(asyncio.CancelledError, Exception):
            await task
        raise RetryableError(display_msg=display_msg) from TimeoutError(
            f"Async operation timed out after {timeout_nanos / 1_000_000_000}s"
        )
    except BaseException:
        if not task.done():
            task.cancel()
            with suppress(asyncio.CancelledError, Exception):
                await task
        raise


#############################################################################
class TestCallWithRetriesAsync(unittest.IsolatedAsyncioTestCase):
    """Unit tests for call_with_retries_async()."""

    async def test_sleep_async_delegates_to_asyncio_sleep(self) -> None:

        sleep_nanos = 456_000_000
        expected_secs = sleep_nanos / 1_000_000_000
        timing = RetryTiming()
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )
        with patch("asyncio.sleep", new=AsyncMock()) as mock_sleep:
            await timing.sleep_async(sleep_nanos, retry)
        mock_sleep.assert_awaited_once_with(expected_secs)

    async def test_call_with_retries_async_success_retries_and_sleeps(self) -> None:
        retry_policy = RetryPolicy(
            max_retries=2,
            min_sleep_secs=0.001,
            initial_max_sleep_secs=0.001,
            max_sleep_secs=0.001,
            max_elapsed_secs=1,
        )
        expected_sleep_nanos: int = 1_000_000
        calls: list[int] = []
        events: list[AttemptOutcome] = []
        sleep_calls: list[tuple[int, int]] = []

        async def fn(retry: Retry) -> str:
            calls.append(retry.count)
            if retry.count < 2:
                raise RetryableError("fail", display_msg="connect") from ValueError("boom")
            return "ok"

        def after_attempt(outcome: AttemptOutcome) -> None:
            events.append(outcome)

        async def fake_sleep_async(sleep_nanos: int, retry: Retry) -> None:
            sleep_calls.append((sleep_nanos, retry.count))

        mock_sleep_async = AsyncMock(side_effect=fake_sleep_async)
        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=mock_sleep_async))
        actual = await call_with_retries_async(
            fn,
            policy=retry_policy,
            after_attempt=after_attempt,
            log=None,
        )

        self.assertEqual("ok", actual)
        self.assertEqual([0, 1, 2], calls)
        self.assertEqual([(expected_sleep_nanos, 0), (expected_sleep_nanos, 1)], sleep_calls)
        self.assertEqual(2, mock_sleep_async.await_count)

        self.assertEqual(3, len(events))
        self.assertFalse(events[0].is_success)
        self.assertFalse(events[0].is_exhausted)
        self.assertEqual(0, events[0].retry.count)
        self.assertEqual(expected_sleep_nanos, events[0].sleep_nanos)

        self.assertFalse(events[1].is_success)
        self.assertFalse(events[1].is_exhausted)
        self.assertEqual(1, events[1].retry.count)
        self.assertEqual(expected_sleep_nanos, events[1].sleep_nanos)

        self.assertTrue(events[2].is_success)
        self.assertEqual(2, events[2].retry.count)
        self.assertEqual(0, events[2].sleep_nanos)

    async def test_call_with_retries_async_retry_immediately_once_skips_backoff_and_sleeps_zero(self) -> None:
        backoff_strategy = MagicMock(side_effect=AssertionError("backoff_strategy must not be called"))
        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0.001,
            initial_max_sleep_secs=0.001,
            max_sleep_secs=0.001,
            max_elapsed_secs=10,
        )
        calls: list[int] = []
        events: list[AttemptOutcome] = []
        sleep_calls: list[tuple[int, int]] = []

        async def fn(retry: Retry) -> str:
            calls.append(retry.count)
            if retry.count == 0:
                raise RetryableError("fail", retry_immediately_once=True) from ValueError("boom")
            return "ok"

        def after_attempt(outcome: AttemptOutcome) -> None:
            events.append(outcome)

        async def fake_sleep_async(sleep_nanos: int, retry: Retry) -> None:
            sleep_calls.append((sleep_nanos, retry.count))

        mock_sleep_async = AsyncMock(side_effect=fake_sleep_async)
        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=mock_sleep_async))
        actual = await call_with_retries_async(
            fn,
            policy=retry_policy,
            backoff=backoff_strategy,
            after_attempt=after_attempt,
            log=None,
        )

        self.assertEqual("ok", actual)
        self.assertEqual([0, 1], calls)
        self.assertEqual([(0, 0)], sleep_calls)
        self.assertEqual(2, len(events))
        self.assertFalse(events[0].is_success)
        self.assertFalse(events[0].is_exhausted)
        self.assertEqual(0, events[0].sleep_nanos)
        self.assertTrue(events[1].is_success)
        self.assertEqual(0, events[1].sleep_nanos)
        self.assertEqual(1, mock_sleep_async.await_count)
        backoff_strategy.assert_not_called()

    async def test_call_with_retries_async_after_attempt_retryable_error_on_success_retries(self) -> None:
        """Ensures raising RetryableError from after_attempt() on a successful attempt triggers a retry."""
        retry_policy = RetryPolicy(max_retries=3, min_sleep_secs=0, initial_max_sleep_secs=0, max_sleep_secs=0)
        fn_calls: list[int] = []
        after_attempt_events: list[tuple[bool, bool, int]] = []
        raised_once: bool = False

        async def fn(retry: Retry) -> str:
            fn_calls.append(retry.count)
            return "ok"

        def after_attempt(outcome: AttemptOutcome) -> None:
            nonlocal raised_once
            after_attempt_events.append((outcome.is_success, outcome.is_exhausted, outcome.retry.count))
            if outcome.is_success and not raised_once:
                raised_once = True
                raise RetryableError("retry from after_attempt")

        actual = await call_with_retries_async(fn, policy=retry_policy, after_attempt=after_attempt, log=None)
        self.assertTrue(raised_once)
        self.assertEqual("ok", actual)
        self.assertEqual([0, 1], fn_calls)
        self.assertEqual([(True, False, 0), (False, False, 0), (True, False, 1)], after_attempt_events)

    async def test_call_with_retries_async_after_attempt_retryable_error_on_failure_aborts(self) -> None:
        """Ensures raising RetryableError from after_attempt() on failure aborts without additional retries."""
        retry_policy = RetryPolicy(max_retries=3, min_sleep_secs=0, initial_max_sleep_secs=0, max_sleep_secs=0)
        fn_calls: list[int] = []
        after_attempt_calls: list[int] = []

        async def fn(retry: Retry) -> None:
            fn_calls.append(retry.count)
            raise RetryableError("attempt failed")

        def after_attempt(outcome: AttemptOutcome) -> None:
            after_attempt_calls.append(outcome.retry.count)
            if not outcome.is_success and not outcome.is_exhausted:
                raise RetryableError("after_attempt abort")

        mock_sleep_async = AsyncMock()
        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=mock_sleep_async))
        with self.assertRaises(RetryableError):
            await call_with_retries_async(fn, policy=retry_policy, after_attempt=after_attempt, log=None)

        self.assertEqual([0], fn_calls)
        self.assertEqual([0], after_attempt_calls)
        mock_sleep_async.assert_not_awaited()

    async def test_call_with_retries_async_giveup_stops_retries(self) -> None:
        """Ensures giveup() stops retries immediately and custom after_attempt disables default logging."""
        retry_policy = RetryPolicy(
            max_retries=5,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        mock_log = MagicMock(spec=logging.Logger)
        calls: list[int] = []

        async def fn(retry: Retry) -> None:
            calls.append(retry.count)
            raise RetryableError("fail") from ValueError("boom")

        def giveup(outcome: AttemptOutcome) -> object | None:
            self.assertEqual(0, outcome.retry.count)
            self.assertEqual(1234, outcome.elapsed_nanos)
            self.assertGreaterEqual(outcome.sleep_nanos, 0)
            self.assertIs(retry_policy, outcome.retry.policy)
            self.assertIsInstance(outcome.result, RetryableError)
            return "circuit breaker triggered"

        after_attempt_events: list[AttemptOutcome] = []

        def after_attempt(outcome: AttemptOutcome) -> None:
            self.assertEqual(0, outcome.retry.count)
            self.assertIsInstance(outcome.result, RetryableError)
            self.assertIs(mock_log, outcome.retry.log)
            self.assertEqual(1234, outcome.elapsed_nanos)
            self.assertGreaterEqual(outcome.sleep_nanos, 0)
            after_attempt_events.append(outcome)

        monotonic_ns = MagicMock(side_effect=[0, 1234])
        retry_policy = retry_policy.copy(timing=RetryTiming(monotonic_ns=monotonic_ns))
        with self.assertRaises(ValueError):
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                giveup=giveup,
                after_attempt=after_attempt,
                log=mock_log,
            )

        # giveup() must prevent additional retries
        self.assertEqual([0], calls)
        mock_log.log.assert_not_called()
        self.assertEqual(1, len(after_attempt_events))
        self.assertFalse(after_attempt_events[0].is_success)
        self.assertTrue(after_attempt_events[0].is_exhausted)
        self.assertFalse(after_attempt_events[0].is_terminated)
        self.assertEqual("circuit breaker triggered", after_attempt_events[0].giveup_reason)

    async def test_call_with_retries_async_before_attempt_can_delay_without_consuming_retry_count(self) -> None:
        """Ensures before_attempt can delay an attempt without incrementing retry.count."""
        retry_policy = RetryPolicy(
            max_retries=0,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
        )
        before_attempt_calls: list[int] = []
        sleeps: list[tuple[int, int]] = []
        fn_calls: list[int] = []

        def before_attempt(retry: Retry) -> int:
            before_attempt_calls.append(retry.count)
            return 123

        async def sleep_async(sleep_nanos: int, retry: Retry) -> None:
            sleeps.append((sleep_nanos, retry.count))

        async def fn(retry: Retry) -> str:
            fn_calls.append(retry.count)
            return "ok"

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))
        self.assertEqual(
            "ok",
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                before_attempt=before_attempt,
                log=None,
            ),
        )
        self.assertEqual([0], before_attempt_calls)
        self.assertEqual([0], fn_calls)
        self.assertEqual([(123, 0)], sleeps)

    async def test_call_with_retries_async_before_attempt_is_invoked_for_each_attempt(self) -> None:
        """Ensures before_attempt runs before each fn(retry) invocation, including retries."""
        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
        )
        before_attempt_calls: list[int] = []
        fn_calls: list[int] = []

        def before_attempt(retry: Retry) -> int:
            before_attempt_calls.append(retry.count)
            return 0 if retry.count == 0 else 7

        sleeps: list[int] = []

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            if sleep_nanos != 0:
                sleeps.append(sleep_nanos)

        async def fn(retry: Retry) -> str:
            fn_calls.append(retry.count)
            if retry.count == 0:
                raise RetryableError("fail", retry_immediately_once=False)
            return "ok"

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))
        self.assertEqual(
            "ok",
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                backoff=lambda ctx: (0, ctx.curr_max_sleep_nanos),
                before_attempt=before_attempt,
                log=None,
            ),
        )
        self.assertEqual([0, 1], before_attempt_calls)
        self.assertEqual([0, 1], fn_calls)
        self.assertEqual([7], sleeps)

    async def test_call_with_retries_async_before_attempt_wrapper_around_noop_reaches_before_attempt_path(self) -> None:
        """Ensures wrapping before_attempt_noop still executes the before_attempt callback path."""
        retry_policy = RetryPolicy.no_retries()
        wrapper_calls: list[int] = []
        sleeps: list[int] = []

        def before_attempt_wrapper(retry: Retry) -> int:
            wrapper_calls.append(retry.count)
            return before_attempt_noop(retry)

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            sleeps.append(sleep_nanos)

        monotonic_ns = MagicMock(side_effect=[1_000, 1_234])
        retry_policy = retry_policy.copy(timing=RetryTiming(monotonic_ns=monotonic_ns).copy(sleep_async=sleep_async))

        async def fn(retry: Retry) -> str:
            self.assertEqual(1_000, retry.call_start_time_nanos)
            self.assertEqual(1_000, retry.before_attempt_start_time_nanos)
            self.assertEqual(1_234, retry.attempt_start_time_nanos)
            self.assertEqual(234, retry.before_attempt_sleep_nanos())
            self.assertEqual(234, retry.idle_nanos)
            return "ok"

        self.assertEqual(
            "ok",
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                before_attempt=before_attempt_wrapper,
                log=None,
            ),
        )
        self.assertEqual([0], wrapper_calls)
        self.assertEqual([], sleeps)
        self.assertEqual(2, monotonic_ns.call_count)

    async def test_call_with_retries_async_before_attempt_negative_sleep_is_an_error(self) -> None:
        """Ensures before_attempt returning a negative duration raises an error."""
        retry_policy = RetryPolicy.no_retries()
        sleeps: list[int] = []

        def before_attempt(_retry: Retry) -> int:
            return -1

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            sleeps.append(sleep_nanos)

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))
        with self.assertRaises(AssertionError):
            await call_with_retries_async(
                lambda _retry: asyncio.sleep(0, result="ok"),
                policy=retry_policy,
                before_attempt=before_attempt,
                log=None,
            )
        self.assertEqual([], sleeps)

    async def test_retry_async_before_attempt_time_nanos_matches_attempt_start_without_sleep(self) -> None:
        """Ensures Retry captures before_attempt_time_nanos when before_attempt returns 0."""
        retry_policy = RetryPolicy.no_retries()

        def before_attempt(_retry: Retry) -> int:
            return 0

        monotonic_ns = MagicMock(side_effect=[1_000, 1_000])
        retry_policy = retry_policy.copy(timing=RetryTiming(monotonic_ns=monotonic_ns).copy(sleep_async=AsyncMock()))

        async def fn(retry: Retry) -> int:
            self.assertEqual(1_000, retry.call_start_time_nanos)
            self.assertEqual(1_000, retry.before_attempt_start_time_nanos)
            self.assertEqual(1_000, retry.attempt_start_time_nanos)
            self.assertEqual(0, retry.before_attempt_sleep_nanos())
            self.assertEqual(0, retry.idle_nanos)
            return 123

        self.assertEqual(
            123,
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                before_attempt=before_attempt,
                log=None,
            ),
        )
        self.assertEqual(2, monotonic_ns.call_count)

    async def test_retry_async_before_attempt_sleep_nanos_reflects_time_elapsed_until_fn(self) -> None:
        """Ensures before_attempt_sleep_nanos is attempt_start_time - before_attempt_time."""
        retry_policy = RetryPolicy.no_retries()

        def before_attempt(_retry: Retry) -> int:
            return 123

        monotonic_ns = MagicMock(side_effect=[1_000, 2_000])
        retry_policy = retry_policy.copy(timing=RetryTiming(monotonic_ns=monotonic_ns).copy(sleep_async=AsyncMock()))

        async def fn(retry: Retry) -> int:
            self.assertEqual(1_000, retry.call_start_time_nanos)
            self.assertEqual(1_000, retry.before_attempt_start_time_nanos)
            self.assertEqual(2_000, retry.attempt_start_time_nanos)
            self.assertEqual(1_000, retry.before_attempt_sleep_nanos())
            self.assertEqual(1_000, retry.idle_nanos)
            return 456

        self.assertEqual(
            456,
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                before_attempt=before_attempt,
                log=None,
            ),
        )

    async def test_retry_async_idle_nanos_includes_retry_sleep_nanos(self) -> None:
        """Ensures Retry.idle_nanos includes retry-loop sleep_nanos on subsequent async attempts."""
        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
        )
        sleeps: list[tuple[int, int]] = []
        retry_idle_nanos: list[tuple[int, int]] = []

        async def sleep_async(sleep_nanos: int, retry: Retry) -> None:
            sleeps.append((sleep_nanos, retry.count))

        retry_policy = retry_policy.copy(timing=RetryTiming(monotonic_ns=lambda: 0).copy(sleep_async=sleep_async))

        async def fn(retry: Retry) -> str:
            retry_idle_nanos.append((retry.count, retry.idle_nanos))
            if retry.count == 0:
                raise RetryableError("fail") from ValueError("boom")
            return "ok"

        self.assertEqual(
            "ok",
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                backoff=lambda ctx: (11, ctx.curr_max_sleep_nanos),
                log=None,
            ),
        )
        self.assertEqual([(11, 0)], sleeps)
        self.assertEqual([(0, 0), (1, 11)], retry_idle_nanos)

    async def test_call_with_retries_async_max_previous_outcomes_1(self) -> None:
        """Ensures Retry.previous_outcomes retains only the most recent AttemptOutcome object."""
        retry_policy = RetryPolicy(
            max_retries=3,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
            max_previous_outcomes=1,
        )
        history_counts_per_attempt: list[list[int]] = []
        nested_history_sizes_per_attempt: list[list[int]] = []

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            self.assertEqual(0, sleep_nanos)

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))

        async def fn(retry: Retry) -> str:
            history_counts_per_attempt.append([outcome.retry.count for outcome in retry.previous_outcomes])
            nested_history_sizes_per_attempt.append(
                [len(outcome.retry.previous_outcomes) for outcome in retry.previous_outcomes]
            )
            for outcome in retry.previous_outcomes:
                self.assertEqual((), outcome.retry.previous_outcomes)
            if retry.count < 2:
                raise RetryableError("fail", retry_immediately_once=(retry.count == 0)) from ValueError("boom")
            return "ok"

        self.assertEqual("ok", await call_with_retries_async(fn, policy=retry_policy, log=None))
        self.assertEqual([[], [0], [1]], history_counts_per_attempt)
        self.assertEqual([[], [0], [0]], nested_history_sizes_per_attempt)

    async def test_call_with_retries_async_max_previous_outcomes_2(self) -> None:
        """Ensures Retry.previous_outcomes retains the last 2 AttemptOutcome objects."""
        retry_policy = RetryPolicy(
            max_retries=3,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
            max_previous_outcomes=2,
        )
        history_counts_per_attempt: list[list[int]] = []

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            self.assertEqual(0, sleep_nanos)

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))

        async def fn(retry: Retry) -> str:
            history_counts_per_attempt.append([outcome.retry.count for outcome in retry.previous_outcomes])
            if retry.count < 3:
                raise RetryableError("fail", retry_immediately_once=(retry.count == 0)) from ValueError("boom")
            return "ok"

        self.assertEqual("ok", await call_with_retries_async(fn, policy=retry_policy, log=None))
        self.assertEqual([[], [0], [0, 1], [1, 2]], history_counts_per_attempt)

    async def test_call_with_retries_async_previous_outcomes_are_detached(self) -> None:
        """Ensures Retry.previous_outcomes entries do not retain their own previous_outcomes history."""
        retry_policy = RetryPolicy(
            max_retries=4,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
            max_previous_outcomes=2,
        )
        nested_history_sizes_per_attempt: list[list[int]] = []

        async def sleep_async(sleep_nanos: int, _retry: Retry) -> None:
            self.assertEqual(0, sleep_nanos)

        retry_policy = retry_policy.copy(timing=RetryTiming().copy(sleep_async=sleep_async))

        async def fn(retry: Retry) -> str:
            nested_history_sizes_per_attempt.append(
                [len(outcome.retry.previous_outcomes) for outcome in retry.previous_outcomes]
            )
            for outcome in retry.previous_outcomes:
                self.assertEqual((), outcome.retry.previous_outcomes)
            if retry.count < 4:
                raise RetryableError("fail", retry_immediately_once=(retry.count == 0)) from ValueError("boom")
            return "ok"

        self.assertEqual("ok", await call_with_retries_async(fn, policy=retry_policy, log=None))
        self.assertEqual([[], [0], [0, 0], [0, 0], [0, 0]], nested_history_sizes_per_attempt)

    async def test_call_with_retries_async_after_attempt_success(self) -> None:
        """Ensures after_attempt is invoked for both retries and final success with correct flags."""
        retry_policy = RetryPolicy(
            max_retries=3,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
        )
        events: list[AttemptOutcome] = []

        async def fn(retry: Retry) -> str:
            # Fail twice, then succeed.
            if retry.count < 2:
                raise RetryableError("fail", retry_immediately_once=True) from ValueError("boom")
            return "ok"

        def after_attempt(outcome: AttemptOutcome) -> None:
            events.append(outcome)
            self.assertFalse(outcome.is_terminated)
            self.assertIsNone(outcome.giveup_reason)
            self.assertIsNone(outcome.retry.log)

        final_result = await call_with_retries_async(
            fn,
            policy=retry_policy,
            after_attempt=after_attempt,
            log=None,
        )
        self.assertEqual("ok", final_result)

        # We expect two failed attempts (counts 0 and 1) and one success (count 2).
        self.assertEqual(3, len(events))

        # All intermediate events must be failures (is_success=False, is_exhausted=False, error present).
        for i, outcome in enumerate(events[:-1]):
            self.assertEqual(i, outcome.retry.count)
            self.assertFalse(outcome.is_success)
            self.assertFalse(outcome.is_exhausted)
            self.assertIsInstance(outcome.result, RetryableError)
            self.assertGreaterEqual(outcome.elapsed_nanos, 0)
            self.assertGreaterEqual(outcome.sleep_nanos, 0)
            if i == 0:
                self.assertEqual(0, outcome.sleep_nanos)

        # Last event must be the success (is_success=True, error is None) for attempt index 2.
        outcome = events[-1]
        self.assertEqual(2, outcome.retry.count)
        self.assertTrue(outcome.is_success)
        self.assertFalse(outcome.is_exhausted)
        self.assertIsNone(outcome.giveup_reason)
        self.assertEqual("ok", outcome.result)
        self.assertGreaterEqual(outcome.elapsed_nanos, 0)
        self.assertEqual(0, outcome.sleep_nanos)

    async def test_call_with_retries_async_after_attempt_exhausted(self) -> None:
        """Ensures after_attempt is invoked with is_exhausted when retries are exhausted."""
        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        events: list[AttemptOutcome] = []

        async def fn(_retry: Retry) -> None:
            raise RetryableError("fail", retry_immediately_once=True) from ValueError("boom")

        def after_attempt(outcome: AttemptOutcome) -> None:
            events.append(outcome)
            self.assertIsNone(outcome.giveup_reason)
            self.assertIsNone(outcome.retry.log)

        with self.assertRaises(ValueError):
            await call_with_retries_async(fn, policy=retry_policy, after_attempt=after_attempt, log=None)

        # There must be at least one event and the last one must indicate exhaustion.
        self.assertGreaterEqual(len(events), 1)
        outcome = events[-1]
        self.assertFalse(outcome.is_success)
        self.assertTrue(outcome.is_exhausted)
        self.assertFalse(outcome.is_terminated)
        self.assertIsNone(outcome.giveup_reason)
        self.assertIsInstance(outcome.result, RetryableError)
        self.assertGreaterEqual(outcome.retry.count, 0)
        self.assertGreaterEqual(outcome.elapsed_nanos, 0)
        self.assertGreaterEqual(outcome.sleep_nanos, 0)

    async def test_call_with_retries_async_can_terminate_after_after_attempt_before_continue(self) -> None:
        """Ensures termination between after_attempt() and the next-attempt continue stops retries."""
        retry_policy = RetryPolicy(
            max_retries=3,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
        )
        calls: list[int] = []
        events: list[AttemptOutcome] = []
        terminated: bool = False

        def is_terminated(_retry: Retry) -> bool:
            return terminated

        retry_policy = retry_policy.copy(timing=RetryTiming(is_terminated=is_terminated))

        async def fn(retry: Retry) -> str:
            calls.append(retry.count)
            raise RetryableError("fail") from ValueError("boom")

        def after_attempt(outcome: AttemptOutcome) -> None:
            nonlocal terminated
            events.append(outcome)
            if not outcome.is_exhausted:
                terminated = True

        actual = await call_with_retries_async(
            fn,
            policy=retry_policy,
            backoff=lambda ctx: (0, ctx.curr_max_sleep_nanos),
            after_attempt=after_attempt,
            on_exhaustion=lambda _outcome: "exhausted",
            log=None,
        )

        self.assertEqual("exhausted", actual)
        self.assertEqual([0], calls)
        self.assertEqual(2, len(events))
        self.assertFalse(events[0].is_exhausted)
        self.assertFalse(events[0].is_terminated)
        self.assertTrue(events[1].is_exhausted)
        self.assertTrue(events[1].is_terminated)
        self.assertEqual(0, events[1].retry.count)

    async def test_call_with_retries_async_on_exhaustion_reraises_cause(self) -> None:
        async def fn(_retry: Retry) -> None:
            raise RetryableError("fail") from ValueError("boom")

        with self.assertRaises(ValueError):
            await call_with_retries_async(fn, policy=RetryPolicy.no_retries(), log=None)

    async def test_call_with_retries_async_on_retryable_error_called_once_per_failure(self) -> None:
        """Ensures on_retryable_error runs once for each RetryableError raised by fn()."""
        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=10,
            reraise=False,
        )
        seen_counts: list[int] = []
        seen_exhausted_flags: list[bool] = []

        def on_retryable_error(outcome: AttemptOutcome) -> None:
            seen_counts.append(outcome.retry.count)
            seen_exhausted_flags.append(outcome.is_exhausted)

        async def fn(_retry: Retry) -> str:
            raise RetryableError("fail") from ValueError("boom")

        with self.assertRaises(RetryError):
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                on_retryable_error=on_retryable_error,
                log=None,
            )

        self.assertEqual([0, 1], seen_counts)
        self.assertEqual([False, False], seen_exhausted_flags)

    async def test_call_with_retries_async_does_not_swallow_cancelled_error(self) -> None:
        """Ensures asyncio.CancelledError propagates immediately and is not treated as a retryable failure.

        In async systems cancellation is used for timeouts and shutdown; swallowing it can hang tasks, block teardown, and
        cause retry loops to ignore cooperative cancellation.
        """
        calls: list[int] = []

        async def fn(retry: Retry) -> None:
            calls.append(retry.count)
            raise asyncio.CancelledError()

        after_attempt = MagicMock()
        on_retryable_error = MagicMock()
        backoff_strategy = MagicMock(side_effect=AssertionError("backoff_strategy must not be called"))
        mock_sleep_async = AsyncMock(side_effect=AssertionError("sleep_async must not be called"))
        retry_policy = RetryPolicy(
            max_retries=10,
            min_sleep_secs=0.001,
            initial_max_sleep_secs=0.001,
            max_sleep_secs=0.001,
            max_elapsed_secs=10,
        ).copy(timing=RetryTiming().copy(sleep_async=mock_sleep_async))

        with self.assertRaises(asyncio.CancelledError):
            await call_with_retries_async(
                fn,
                policy=retry_policy,
                backoff=backoff_strategy,
                after_attempt=after_attempt,
                on_retryable_error=on_retryable_error,
                log=None,
            )

        self.assertEqual([0], calls)
        after_attempt.assert_not_called()
        on_retryable_error.assert_not_called()
        backoff_strategy.assert_not_called()
        self.assertEqual(0, mock_sleep_async.await_count)

    async def test_retry_template_call_with_retries_async(self) -> None:
        """Ensures RetryTemplate.call_with_retries_async() runs with template defaults and log overrides."""
        calls: list[int] = []
        before_attempt_calls: list[int] = []
        outcomes: list[AttemptOutcome] = []
        retry_policy = RetryPolicy.no_retries()
        retry_config = RetryConfig(display_msg="template")
        retry_policy = retry_policy.copy(config=retry_config)
        template_log = MagicMock(spec=logging.Logger)

        def before_attempt(retry: Retry) -> int:
            before_attempt_calls.append(retry.count)
            return 0

        def after_attempt(outcome: AttemptOutcome) -> None:
            outcomes.append(outcome)

        template: RetryTemplate[str] = RetryTemplate(
            policy=retry_policy,
            before_attempt=before_attempt,
            after_attempt=after_attempt,
            log=template_log,
        )

        async def fn(retry: Retry) -> str:
            calls.append(retry.count)
            self.assertIs(retry_policy, retry.policy)
            self.assertIs(retry_config, retry.policy.config)
            return "hello"

        with self.subTest("default_logger_from_template"):
            calls.clear()
            before_attempt_calls.clear()
            outcomes.clear()
            actual = await template.call_with_retries_async(fn)
            self.assertEqual("hello", actual)
            self.assertEqual([0], calls)
            self.assertEqual([0], before_attempt_calls)
            self.assertEqual(1, len(outcomes))
            self.assertTrue(outcomes[0].is_success)
            self.assertIs(template_log, outcomes[0].retry.log)

        with self.subTest("no_logger_override"):
            calls.clear()
            before_attempt_calls.clear()
            outcomes.clear()
            actual = await template.call_with_retries_async(fn, log=NO_LOGGER)
            self.assertEqual("hello", actual)
            self.assertEqual([0], calls)
            self.assertEqual([0], before_attempt_calls)
            self.assertEqual(1, len(outcomes))
            self.assertTrue(outcomes[0].is_success)
            self.assertIsNone(outcomes[0].retry.log)

    async def test_await_with_timeout_success_and_timeout(self) -> None:
        loop = asyncio.get_running_loop()

        with self.subTest("success"):

            async def immediate() -> str:
                return "ok"

            actual = await asyncio_await_with_timeout(immediate(), timeout_nanos=1_000_000_000)
            self.assertEqual("ok", actual)

        with self.subTest("timeout"):
            cancelled = asyncio.Event()

            async def long_running() -> None:
                try:
                    await asyncio.sleep(1)
                except asyncio.CancelledError:
                    cancelled.set()
                    raise

            with self.assertRaises(RetryableError) as exc:
                await asyncio_await_with_timeout(long_running(), timeout_nanos=50_000_000, display_msg="connect")
            self.assertEqual("connect", exc.exception.display_msg)
            self.assertIsInstance(exc.exception.__cause__, TimeoutError)
            self.assertTrue(cancelled.is_set())

        with self.subTest("invalid_timeout"):
            fut: asyncio.Future[None] = loop.create_future()
            fut.set_result(None)
            with self.assertRaises(ValueError):
                await asyncio_await_with_timeout(fut, timeout_nanos=-1)

        with self.subTest("zero_timeout"):
            fut2: asyncio.Future[str] = loop.create_future()
            fut2.set_result("ok")
            with self.assertRaises(RetryableError) as exc2:
                await asyncio_await_with_timeout(fut2, timeout_nanos=0, display_msg="connect")
            self.assertEqual("connect", exc2.exception.display_msg)
            self.assertIsInstance(exc2.exception.__cause__, TimeoutError)

    async def test_make_from_asyncio_none_returns_default_timing(self) -> None:
        sleep_nanos = 456_000_000
        expected_secs = sleep_nanos / 1_000_000_000
        timing = RetryTiming.make_from_asyncio(None)
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )
        with patch("asyncio.sleep", new=AsyncMock()) as mock_sleep:
            await timing.sleep_async(sleep_nanos, retry)
        mock_sleep.assert_awaited_once_with(expected_secs)
        self.assertFalse(timing.is_terminated(retry))

    async def test_make_from_asyncio_zero_sleep_yields(self) -> None:
        termination_event = asyncio.Event()
        timing = RetryTiming.make_from_asyncio(termination_event)
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )
        with (
            patch("asyncio.sleep", new=AsyncMock()) as mock_sleep,
            patch("asyncio.wait_for", new=AsyncMock()) as mock_wait_for,
        ):
            await timing.sleep_async(0, retry)
        mock_sleep.assert_awaited_once_with(0)
        mock_wait_for.assert_not_awaited()

    def test_make_from_asyncio_event_is_terminated_delegates_to_is_set(self) -> None:
        termination_event = MagicMock(spec=asyncio.Event)
        termination_event.is_set.return_value = True
        timing = RetryTiming.make_from_asyncio(termination_event)
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )
        self.assertTrue(timing.is_terminated(retry))
        termination_event.is_set.assert_called_once()

    async def test_make_from_asyncio_positive_sleep_waits_for_termination_event(self) -> None:
        sleep_nanos = 456_000_000
        expected_secs = sleep_nanos / 1_000_000_000
        termination_event = MagicMock()
        termination_event.is_set.return_value = False
        fut: asyncio.Future[None] = asyncio.get_running_loop().create_future()
        fut.set_result(None)
        termination_event.wait.return_value = fut
        timing = RetryTiming.make_from_asyncio(termination_event)
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )

        async def fake_wait_for(awaitable: Awaitable[object], *, timeout: float) -> None:
            self.assertEqual(expected_secs, timeout)
            await awaitable

        with (
            patch("asyncio.wait_for", new=AsyncMock(side_effect=fake_wait_for)) as mock_wait_for,
            patch("asyncio.sleep", new=AsyncMock()) as mock_sleep,
        ):
            await timing.sleep_async(sleep_nanos, retry)

        termination_event.wait.assert_called_once_with()
        mock_wait_for.assert_awaited_once()
        self.assertIs(fut, mock_wait_for.call_args[0][0])
        self.assertEqual(expected_secs, mock_wait_for.call_args.kwargs["timeout"])
        mock_sleep.assert_not_awaited()

    async def test_make_from_asyncio_positive_sleep_timeout_is_swallowed(self) -> None:
        sleep_nanos = 456_000_000
        expected_secs = sleep_nanos / 1_000_000_000
        termination_event = MagicMock()
        termination_event.is_set.return_value = False
        fut: asyncio.Future[None] = asyncio.get_running_loop().create_future()
        termination_event.wait.return_value = fut
        timing = RetryTiming.make_from_asyncio(termination_event)
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=timing),
            log=None,
            previous_outcomes=(),
        )

        mock_wait_for = AsyncMock(side_effect=asyncio.TimeoutError)
        with (
            patch("asyncio.wait_for", new=mock_wait_for),
            patch("asyncio.sleep", new=AsyncMock()) as mock_sleep,
        ):
            await timing.sleep_async(sleep_nanos, retry)

        termination_event.wait.assert_called_once_with()
        mock_wait_for.assert_awaited_once()
        self.assertIs(fut, mock_wait_for.call_args[0][0])
        self.assertEqual(expected_secs, mock_wait_for.call_args.kwargs["timeout"])
        mock_sleep.assert_not_awaited()

    def test_make_from_asyncio_none_result_is_pickleable(self) -> None:
        """RetryTiming.make_from_asyncio(None) must remain pickleable."""
        timing = RetryTiming.make_from_asyncio(None)
        roundtripped = pickle.loads(pickle.dumps(timing))
        retry = Retry(
            count=0,
            call_start_time_nanos=0,
            before_attempt_start_time_nanos=0,
            attempt_start_time_nanos=0,
            idle_nanos=0,
            policy=RetryPolicy.no_retries().copy(timing=roundtripped),
            log=None,
            previous_outcomes=(),
        )
        self.assertFalse(roundtripped.is_terminated(retry))


#############################################################################
class TestRetryTemplateWrapsAsync(unittest.IsolatedAsyncioTestCase):

    async def test_wraps_async_function_retries_and_passes_args(self) -> None:
        calls: list[tuple[int, int]] = []
        retry_count = 0

        async def fn(x: int) -> int:
            nonlocal retry_count
            calls.append((retry_count, x))
            retry_count += 1
            if retry_count == 1:
                raise RetryableError("fail")
            return x * 2

        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        template: RetryTemplate = RetryTemplate(policy=retry_policy, log=None)
        wrapped: Callable[[int], Awaitable[int]] = template.wraps(fn)

        actual: int = await wrapped(5)
        self.assertEqual(10, actual)
        self.assertEqual([(0, 5), (1, 5)], calls)

    async def test_wraps_async_callable_object_retries_and_passes_args(self) -> None:
        """Ensures wraps() treats async __call__ objects as async retry targets."""
        calls: list[tuple[int, int]] = []

        class MyAsyncCallable:
            """Async callable that fails once, then succeeds."""

            retry_count = 0

            async def __call__(self, x: int) -> int:
                """Raises RetryableError on first call and returns x*2 thereafter."""
                calls.append((self.retry_count, x))
                self.retry_count += 1
                if self.retry_count == 1:
                    raise RetryableError("fail")
                return x * 2

        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        template: RetryTemplate[int] = RetryTemplate(policy=retry_policy, log=None)
        wrapped: Callable[[int], Awaitable[int]] = template.wraps(MyAsyncCallable())

        actual: int = await wrapped(5)
        self.assertEqual(10, actual)
        self.assertEqual([(0, 5), (1, 5)], calls)

    async def test_wraps_partial_of_async_callable_object_retries_and_passes_args(self) -> None:
        """Ensures wraps() handles functools.partial(async_callable_obj, ...) correctly."""
        calls: list[tuple[int, int]] = []

        class MyAsyncCallable:
            """Async callable that fails once, then succeeds."""

            retry_count = 0

            async def __call__(self, x: int) -> int:
                """Raises RetryableError on first call and returns x*2 thereafter."""
                calls.append((self.retry_count, x))
                self.retry_count += 1
                if self.retry_count == 1:
                    raise RetryableError("fail")
                return x * 2

        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        template: RetryTemplate = RetryTemplate(policy=retry_policy, log=None)
        wrapped: Callable[[], Awaitable[int]] = template.wraps(functools.partial(MyAsyncCallable(), 5))

        actual: int = await wrapped()
        self.assertEqual(10, actual)
        self.assertEqual([(0, 5), (1, 5)], calls)

    async def test_wraps_partial_of_async_function_retries_and_passes_args(self) -> None:
        """Ensures wraps() handles functools.partial(async_fn, ...) correctly."""
        calls: list[tuple[int, int]] = []
        retry_count = 0

        async def fn(x: int) -> int:
            nonlocal retry_count
            calls.append((retry_count, x))
            retry_count += 1
            if retry_count == 1:
                raise RetryableError("fail")
            return x * 2

        retry_policy = RetryPolicy(
            max_retries=1,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1,
        )
        template: RetryTemplate = RetryTemplate(policy=retry_policy, log=None)
        wrapped: Callable[[], Awaitable[int]] = template.wraps(functools.partial(fn, 5))

        actual: int = await wrapped()
        self.assertEqual(10, actual)
        self.assertEqual([(0, 5), (1, 5)], calls)



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Shared environment and base classes for integration tests.

Holds the integration test globals (e.g., temp ZFS key material, pool/dataset state) and common helpers so that split-out
integration test modules can share a single runtime context.
"""

from __future__ import (
    annotations,
)
import logging
import os
import platform
import pwd
import random
import re
import shutil
import subprocess
import tempfile
import threading
import traceback
import unittest
from collections import (
    Counter,
)
from collections.abc import (
    Iterable,
)
from pathlib import (
    Path,
)
from subprocess import (
    PIPE,
)
from typing import (
    Any,
    Callable,
    Final,
    cast,
)

from bzfs_main import (
    argparse_cli,
    bzfs,
    bzfs_jobrunner,
)
from bzfs_main.detect import (
    is_version_at_least,
)
from bzfs_main.replication import (
    INJECT_DST_PIPE_FAIL_KBYTES,
)
from bzfs_main.util.utils import (
    ENV_VAR_PREFIX,
    FILE_PERMISSIONS,
    find_match,
    get_home_directory,
    getenv_any,
    getenv_bool,
)
from bzfs_tests.abstract_testcase import (
    AbstractTestCase,
)
from bzfs_tests.zfs_util import (
    bookmark_name,
    bookmarks,
    build,
    create_filesystem,
    create_volume,
    dataset_exists,
    dataset_property,
    destroy,
    destroy_pool,
    run_cmd,
    set_sudo_cmd,
    snapshot_name,
    snapshots,
    take_snapshot,
    zfs_version,
)

# constants:
SRC_POOL_NAME: Final[str] = "wb_src"
DST_POOL_NAME: Final[str] = "wb_dst"
POOL_SIZE_BYTES_DEFAULT: Final[int] = 100 * 1024 * 1024
ENCRYPTION_ALGO: Final[str] = "aes-256-gcm"
CREATION_PREFIX: Final[str] = "bzfs_test:"

# Global variables populated during setUp()
SRC_POOL: str = ""
DST_POOL: str = ""
SRC_ROOT_DATASET: str = ""
DST_ROOT_DATASET: str = ""
AFIX: str = ""
ZPOOL_FEATURES: dict[str, dict[str, str]] | None = None

ZFS_ENCRYPTIONK_KEY_FD, ZFS_ENCRYPTION_KEY = tempfile.mkstemp(prefix="test_bzfs.key_")
os.write(ZFS_ENCRYPTIONK_KEY_FD, "mypasswd".encode("utf-8"))
os.close(ZFS_ENCRYPTIONK_KEY_FD)
KEYLOCATION: Final[str] = f"file://{ZFS_ENCRYPTION_KEY}"

SSH_CONFIG_FILE: Final[str] = os.path.join(get_home_directory(), ".ssh", "test_bzfs_ssh_config")
RNG: Final[random.Random] = random.Random(12345)

SSH_PROGRAM: Final[str] = cast(str, getenv_any("test_ssh_program", "ssh"))  # also works with "hpnssh"
SUDO_CMD: Final[list[str]] = ["sudo", "-n"] if getenv_bool("test_enable_sudo", True) and os.getuid() != 0 else []
set_sudo_cmd(SUDO_CMD)


#############################################################################
class ParametrizedTestCase(AbstractTestCase):

    def __init__(self, methodName: str = "runTest", param: dict[str, Any] | None = None) -> None:  # noqa: N803
        super().__init__(methodName)
        self.param = param

    @staticmethod
    def parametrize(testcase_klass: type, param: dict[str, Any] | None = None) -> unittest.TestSuite:
        testloader = unittest.TestLoader()
        testnames = testloader.getTestCaseNames(testcase_klass)
        suite = unittest.TestSuite()
        for name in testnames:
            suite.addTest(testcase_klass(name, param=param))
        return suite


#############################################################################
class IntegrationTestCase(ParametrizedTestCase):

    def setUp(self, pool_size_bytes: int = POOL_SIZE_BYTES_DEFAULT) -> None:
        global SRC_POOL, DST_POOL
        global SRC_ROOT_DATASET, DST_ROOT_DATASET
        global AFIX

        logdir = self.log_dir_opt()[-1]
        if os.path.exists(logdir):
            shutil.rmtree(logdir)
        for pool in SRC_POOL_NAME, DST_POOL_NAME:
            if dataset_exists(pool):
                destroy_pool(pool)
            if not dataset_exists(pool):
                tmp = tempfile.NamedTemporaryFile()  # noqa: SIM115
                tmp.seek(pool_size_bytes - 1)
                tmp.write(b"0")
                tmp.seek(0)
                run_cmd(SUDO_CMD + ["zpool", "create", "-O", "atime=off", pool, tmp.name])

        SRC_POOL = build(SRC_POOL_NAME)
        DST_POOL = build(DST_POOL_NAME)
        AFIX = self.param.get("affix", "") if self.param is not None else ""
        SRC_ROOT_DATASET = recreate_filesystem(SRC_POOL_NAME + "/tmp/" + fix("src"))
        DST_ROOT_DATASET = recreate_filesystem(DST_POOL_NAME + "/tmp/" + fix("dst"))

        global ZPOOL_FEATURES
        if ZPOOL_FEATURES is None:
            ZPOOL_FEATURES = {}
            detect_zpool_features("src", SRC_POOL_NAME)
            logging.getLogger(__name__).warning(f"zpool bookmarks feature: {are_bookmarks_enabled('src')}")
            props = ZPOOL_FEATURES["src"]
            features = "\n".join(
                [f"{k}: {v}" for k, v in sorted(props.items()) if k.startswith("feature@") or k == "delegation"]
            )
            self.assertIsNotNone(features)
            # print(f"test zpool features: {features}", file=sys.stderr)

        private_key_file = "id_rsa" if platform.system() == "Linux" else "testid_rsa"
        private_key_file = os.path.join(os.path.dirname(SSH_CONFIG_FILE), private_key_file)
        fallback = "Fallback=no\n" if SSH_PROGRAM == "hpnssh" else ""  # see https://www.psc.edu/hpn-ssh-home/hpn-readme/
        with open(SSH_CONFIG_FILE, "w", encoding="utf-8") as fd:
            fd.write(f"IdentityFile {private_key_file}\nStrictHostKeyChecking=no\n{fallback}Include /etc/ssh/ssh_config\n")
        os.chmod(SSH_CONFIG_FILE, mode=FILE_PERMISSIONS)  # chmod u=rw,go=

    # zpool list -o name|grep '^wb_'|xargs -n 1 -r --verbose zpool destroy; rm -fr /tmp/tmp* /run/user/$UID/bzfs/
    def tearDown(self) -> None:
        Path(SSH_CONFIG_FILE).unlink(missing_ok=True)
        # tear down is deferred to next setup for easier debugging

    def tearDownAndSetup(self) -> None:  # noqa: N802
        self.tearDown()
        self.setUp()

    def setup_basic(self, volume: bool = False) -> None:
        compression_props = ["-o", "compression=on"]
        encryption_props = ["-o", f"encryption={ENCRYPTION_ALGO}"]
        encryption_props += ["-o", "keyformat=passphrase", "-o", f"keylocation={KEYLOCATION}"]

        dataset_props = encryption_props + compression_props if self.is_encryption_mode() else compression_props
        src_foo = create_filesystem(SRC_ROOT_DATASET, "foo", props=dataset_props)
        src_foo_a = create_volume(src_foo, "a", size="1M") if volume else create_filesystem(src_foo, "a")
        src_foo_b = create_filesystem(src_foo, "b")
        self.assertIsNotNone(src_foo_b)
        take_snapshot(SRC_ROOT_DATASET, fix("s1"))
        take_snapshot(SRC_ROOT_DATASET, fix("s2"))
        take_snapshot(SRC_ROOT_DATASET, fix("s3"))
        take_snapshot(src_foo, fix("t1"))
        take_snapshot(src_foo, fix("t2"))
        take_snapshot(src_foo, fix("t3"))
        take_snapshot(src_foo_a, fix("u1"))
        take_snapshot(src_foo_a, fix("u2"))
        take_snapshot(src_foo_a, fix("u3"))

    def setup_basic_woo(self, w: int = 1, q: int = 0) -> None:
        for i in range(w):
            src_woo = create_filesystem(SRC_ROOT_DATASET, f"woo{i}")
            take_snapshot(src_woo, fix("w1"))
            take_snapshot(src_woo, fix("w2"))
            take_snapshot(src_woo, fix("w3"))
            for j in range(q):
                src_woo_qoo = create_filesystem(src_woo, f"qoo{j}")
                take_snapshot(src_woo_qoo, fix("q1"))
                take_snapshot(src_woo_qoo, fix("q2"))
                take_snapshot(src_woo_qoo, fix("q3"))

    def setup_basic_with_recursive_replication_done(self) -> None:
        self.setup_basic()
        self.run_bzfs(SRC_ROOT_DATASET, DST_ROOT_DATASET, "--recursive")
        self.assert_snapshots(DST_ROOT_DATASET, 3, "s")
        self.assert_snapshots(DST_ROOT_DATASET + "/foo", 3, "t")
        self.assert_snapshots(DST_ROOT_DATASET + "/foo/a", 3, "u")
        self.assertFalse(dataset_exists(DST_ROOT_DATASET + "/foo/b"))  # b/c src has no snapshots

    @staticmethod
    def log_dir_opt() -> list[str]:
        return ["--log-dir", os.path.join(get_home_directory(), argparse_cli.LOG_DIR_DEFAULT + "-test")]

    def run_bzfs_internal(
        self,
        *arguments: str,
        dry_run: bool | None = None,
        no_create_bookmark: bool = False,
        no_use_bookmark: bool = False,
        skip_on_error: str = "fail",
        retries: int = 0,
        expected_status: int | list[int] = 0,
        error_injection_triggers: dict[str, Counter] | None = None,
        delete_injection_triggers: dict[str, Counter] | None = None,
        param_injection_triggers: dict[str, dict[str, bool]] | None = None,
        inject_params: dict[str, bool] | None = None,
        max_command_line_bytes: int | None = None,
        creation_prefix: str | None = None,
        max_exceptions_to_summarize: int | None = None,
        max_datasets_per_minibatch_on_list_snaps: int | None = None,
        ssh_control_persist_margin_secs: int | None = None,
        isatty: bool | None = None,
        progress_update_intervals: tuple[float, float] | None = None,
        use_select: bool | None = None,
        use_jobrunner: bool = False,
        spawn_process_per_job: bool | None = None,
        cache_snapshots: bool = False,
        force_stable_cache_namespace: bool = False,
    ) -> bzfs.Job | bzfs_jobrunner.Job:
        port = getenv_any("test_ssh_port")  # set this if sshd is on non-standard port: export bzfs_test_ssh_port=12345
        args = list(arguments)
        src_host = [] if use_jobrunner else ["--ssh-src-host", "127.0.0.1"]
        dst_host = [] if use_jobrunner else ["--ssh-dst-host", "127.0.0.1"]
        ssh_dflt_port = "2222" if SSH_PROGRAM == "hpnssh" else "22"  # see https://www.psc.edu/hpn-ssh-home/hpn-readme/
        src_port = ["--ssh-src-port", ssh_dflt_port if port is None else str(port)]
        dst_port = [] if port is None else ["--ssh-dst-port", str(port)]
        src_user = ["--ssh-src-user", os_username()]
        params = self.param
        if params and params.get("ssh_mode") == "push":
            args += dst_host + dst_port
        elif params and params.get("ssh_mode") == "pull":
            args += src_host + src_port
        elif params and params.get("ssh_mode") == "pull-push":
            if (
                getenv_bool("test_enable_IPv6", True)
                and RNG.randint(0, 2) % 3 == 0
                and not platform.platform().startswith("FreeBSD-")
                and not SSH_PROGRAM == "hpnssh"
                and not force_stable_cache_namespace
            ):
                src_host = [] if use_jobrunner else ["--ssh-src-host", "::1"]  # IPv6 syntax for 127.0.0.1 loopback address
                dst_host = [] if use_jobrunner else ["--ssh-dst-host", "::1"]  # IPv6 syntax for 127.0.0.1 loopback address
            args += src_host + dst_host + src_port + dst_port
            if params and "min_pipe_transfer_size" in params and int(params["min_pipe_transfer_size"]) == 0:
                args += src_user
            args += ["--bwlimit=10000m"]
        elif params and params.get("ssh_mode", "local") != "local":
            raise ValueError("Unknown ssh_mode: " + params["ssh_mode"])

        args += [
            "--ssh-src-config-file=" + SSH_CONFIG_FILE,
            "--ssh-dst-config-file=" + SSH_CONFIG_FILE,
        ]

        for arg in args:
            assert "--retries" not in arg
        args += [f"--retries={retries}"]
        args += self.log_dir_opt()

        if params and "skip_missing_snapshots" in params:
            i = find_match(args, lambda arg: arg.startswith("-"))
            i = max(i, 0)
            args = args[0:i] + ["--skip-missing-snapshots=" + str(params["skip_missing_snapshots"])] + args[i:]

        if self.is_no_privilege_elevation():
            # test ZFS delegation in combination with --no-privilege-elevation flag
            args = args + ["--no-privilege-elevation"]
            src_permissions = "send,snapshot,hold,bookmark"
            if delete_injection_triggers is not None:
                src_permissions += ",destroy,mount"
            # optional_dst_permissions = ",canmount,mountpoint,readonly,compression,encryption,keylocation,recordsize"
            optional_dst_permissions = ",keylocation,compression"
            dst_permissions = "mount,create,receive,rollback,destroy" + optional_dst_permissions
            cmd = f"sudo -n zfs allow -u {os_username()} {src_permissions}".split(" ") + [SRC_POOL_NAME]
            if dataset_exists(SRC_POOL_NAME):
                run_cmd(cmd)
            cmd = f"sudo -n zfs allow -u {os_username()} {dst_permissions}".split(" ") + [DST_POOL_NAME]
            if dataset_exists(DST_POOL_NAME):
                run_cmd(cmd)

        if SSH_PROGRAM != "ssh" and "--ssh-program" not in args and "--ssh-program=" not in args:
            args = args + ["--ssh-program=" + SSH_PROGRAM]

        if params and params.get("verbose", None):
            args = args + ["--verbose"]

        if params and "min_pipe_transfer_size" in params:
            old_min_pipe_transfer_size = os.environ.get(ENV_VAR_PREFIX + "min_pipe_transfer_size")
            os.environ[ENV_VAR_PREFIX + "min_pipe_transfer_size"] = str(int(params["min_pipe_transfer_size"]))

        if dry_run:
            args = args + ["--dryrun=recv"]

        if no_create_bookmark:
            args = args + ["--create-bookmarks=none"]

        if no_use_bookmark:
            args = args + ["--no-use-bookmark"]

        if skip_on_error:
            args = args + ["--skip-on-error=" + skip_on_error]

        args = args + ["--exclude-envvar-regex=EDITOR"]
        args += ["--cache-snapshots"] if cache_snapshots else []

        job: bzfs.Job | bzfs_jobrunner.Job
        if use_jobrunner:
            job = bzfs_jobrunner.Job(log=None, termination_event=threading.Event())
            job.is_test_mode = True
            if spawn_process_per_job:
                args += ["--spawn-process-per-job"]
        else:
            job = bzfs.Job()
            job.is_test_mode = True

            if error_injection_triggers is not None:
                job.error_injection_triggers = error_injection_triggers
                args = args + ["--threads=1"]

            if delete_injection_triggers is not None:
                job.delete_injection_triggers = delete_injection_triggers
                args = args + ["--threads=1"]

            if param_injection_triggers is not None:
                job.param_injection_triggers = param_injection_triggers
                args = args + ["--threads=1"]

            if inject_params is not None:
                job.inject_params = inject_params

            if max_command_line_bytes is not None:
                job.max_command_line_bytes = max_command_line_bytes

            if creation_prefix is not None:
                job.creation_prefix = creation_prefix

            if max_exceptions_to_summarize is not None:
                job.max_exceptions_to_summarize = max_exceptions_to_summarize

            if use_select is not None:
                job.use_select = use_select

            if progress_update_intervals is not None:
                job.progress_update_intervals = progress_update_intervals

        old_ssh_control_persist_margin_secs = os.environ.get(ENV_VAR_PREFIX + "ssh_control_persist_margin_secs")
        if ssh_control_persist_margin_secs is not None:
            os.environ[ENV_VAR_PREFIX + "ssh_control_persist_margin_secs"] = str(ssh_control_persist_margin_secs)

        old_max_datasets_per_minibatch_on_list_snaps = os.environ.get(
            ENV_VAR_PREFIX + "max_datasets_per_minibatch_on_list_snaps"
        )
        if max_datasets_per_minibatch_on_list_snaps is not None:
            os.environ[ENV_VAR_PREFIX + "max_datasets_per_minibatch_on_list_snaps"] = str(
                max_datasets_per_minibatch_on_list_snaps
            )

        old_dedicated_tcp_connection_per_zfs_send = os.environ.get(ENV_VAR_PREFIX + "dedicated_tcp_connection_per_zfs_send")
        if platform.platform().startswith("FreeBSD-13"):
            # workaround for spurious hangs during zfs send/receive in ~30% of Github Action jobs on QEMU
            # probably caused by https://bugs.freebsd.org/bugzilla/show_bug.cgi?id=283101
            # via https://github.com/openzfs/zfs/issues/16731#issuecomment-2561987688
            os.environ[ENV_VAR_PREFIX + "dedicated_tcp_connection_per_zfs_send"] = "false"

        old_isatty = os.environ.get(ENV_VAR_PREFIX + "isatty")
        if isatty is not None:
            os.environ[ENV_VAR_PREFIX + "isatty"] = str(isatty)

        returncode = 0
        try:
            if use_jobrunner:
                assert isinstance(job, bzfs_jobrunner.Job)
                job.run_main([bzfs_jobrunner.PROG_NAME] + args)
            else:
                assert isinstance(job, bzfs.Job)
                job.run_main(bzfs.argument_parser().parse_args(args), args)
        except subprocess.CalledProcessError as e:
            returncode = e.returncode
            if expected_status != returncode:
                traceback.print_exc()
        except SystemExit as e:
            assert isinstance(e.code, int)
            returncode = e.code
            if expected_status != returncode:
                traceback.print_exc()
        finally:
            if self.is_no_privilege_elevation():
                # revoke all ZFS delegation permissions
                cmd = f"sudo -n zfs unallow -r -u {os_username()}".split(" ") + [SRC_POOL_NAME]
                if dataset_exists(SRC_POOL_NAME):
                    run_cmd(cmd)
                cmd = f"sudo -n zfs unallow -r -u {os_username()}".split(" ") + [DST_POOL_NAME]
                if dataset_exists(DST_POOL_NAME):
                    run_cmd(cmd)

            if params and "min_pipe_transfer_size" in params:
                if old_min_pipe_transfer_size is None:
                    os.environ.pop(ENV_VAR_PREFIX + "min_pipe_transfer_size", None)
                else:
                    os.environ[ENV_VAR_PREFIX + "min_pipe_transfer_size"] = old_min_pipe_transfer_size

            if max_datasets_per_minibatch_on_list_snaps is not None:
                if old_max_datasets_per_minibatch_on_list_snaps is None:
                    os.environ.pop(ENV_VAR_PREFIX + "max_datasets_per_minibatch_on_list_snaps", None)
                else:
                    os.environ[ENV_VAR_PREFIX + "max_datasets_per_minibatch_on_list_snaps"] = (
                        old_max_datasets_per_minibatch_on_list_snaps
                    )

            if old_dedicated_tcp_connection_per_zfs_send is None:
                os.environ.pop(ENV_VAR_PREFIX + "dedicated_tcp_connection_per_zfs_send", None)
            else:
                os.environ[ENV_VAR_PREFIX + "dedicated_tcp_connection_per_zfs_send"] = (
                    old_dedicated_tcp_connection_per_zfs_send
                )

            if old_isatty is None:
                os.environ.pop(ENV_VAR_PREFIX + "isatty", None)
            else:
                os.environ[ENV_VAR_PREFIX + "isatty"] = old_isatty

            if old_ssh_control_persist_margin_secs is None:
                os.environ.pop(ENV_VAR_PREFIX + "ssh_control_persist_margin_secs", None)
            else:
                os.environ[ENV_VAR_PREFIX + "ssh_control_persist_margin_secs"] = old_ssh_control_persist_margin_secs

        if isinstance(expected_status, list):
            self.assertIn(returncode, expected_status)
        else:
            self.assertEqual(expected_status, returncode)
        return job

    def run_bzfs(self, *args: str, **kwargs: Any) -> bzfs.Job:
        job = self.run_bzfs_internal(*args, **kwargs)
        assert isinstance(job, bzfs.Job)
        return job

    def run_bzfs_jobrunner(self, *args: str, **kwargs: Any) -> bzfs_jobrunner.Job:
        job = self.run_bzfs_internal(*args, use_jobrunner=True, **kwargs)
        assert isinstance(job, bzfs_jobrunner.Job)
        return job

    def assert_snapshot_names(self, dataset: str, expected_names: list[str]) -> None:
        dataset = build(dataset)
        snap_names = natsorted([snapshot_name(snapshot) for snapshot in snapshots(dataset)])
        expected_names = [fix(name) for name in expected_names]
        self.assertListEqual(expected_names, snap_names)

    def assert_snapshots(
        self, dataset: str, expected_num_snapshots: int, snapshot_prefix: str = "", offset: int = 0
    ) -> None:
        expected_names = [f"{snapshot_prefix}{i + 1 + offset}" for i in range(expected_num_snapshots)]
        self.assert_snapshot_names(dataset, expected_names)

    def assert_snapshot_name_regexes(self, dataset: str, expected_names: list[str]) -> None:
        dataset = build(dataset)
        snap_names = natsorted([snapshot_name(snapshot) for snapshot in snapshots(dataset)])
        expected_names = [fix(name) for name in expected_names]
        self.assertEqual(len(expected_names), len(snap_names), f"{expected_names} vs. {snap_names}")
        for expected_name, snap_name in zip(expected_names, snap_names):
            self.assertRegex(snap_name, expected_name, f"{expected_names} vs. {snap_names}")

    def assert_bookmark_names(self, dataset: str, expected_names: list[str]) -> None:
        dataset = build(dataset)
        snap_names = natsorted([bookmark_name(bookmark) for bookmark in bookmarks(dataset)])
        expected_names = [fix(name) for name in expected_names]
        self.assertListEqual(expected_names, snap_names)

    def is_no_privilege_elevation(self) -> bool:
        return bool(self.param and self.param.get("no_privilege_elevation", False))

    def is_encryption_mode(self) -> bool:
        return bool(self.param and self.param.get("encrypted_dataset", False))

    @staticmethod
    def properties_with_special_characters() -> dict[str, str]:
        return {
            "compression": "off",
            "bzfs:prop0": "/tmp/dir with  spaces and $dollar sign-" + str(os.getpid()),
            "bzfs:prop1": "/tmp/dir` ~!@#$%^&*()_+-={}[]|;:<>?,./",  # test escaping
            "bzfs:prop2": "/tmp/foo'bar",
            "bzfs:prop3": '/tmp/foo"bar',
            "bzfs:prop4": "/tmp/foo'ba\"rbaz",
            "bzfs:prop5": '/tmp/foo"ba\'r"baz',
            "bzfs:prop6": "/tmp/foo  bar\t\t\nbaz\n\n\n",
            "bzfs:prop7": "/tmp/foo\\bar",
            "bzfs:prop8": "\\\\server\\share",  # UNC-style path with leading double backslashes
            "bzfs:prop9": "/tmp/foo\\`bar",  # backslash + backtick sequence exercises dquote edge case
        }

    def generate_recv_resume_token(self, from_snapshot: str | None, to_snapshot: str, dst_dataset: str) -> None:
        snapshot_opts = to_snapshot if not from_snapshot else f"-i {from_snapshot} {to_snapshot}"
        send = f"sudo -n zfs send --raw --compressed -v {snapshot_opts}"
        c = INJECT_DST_PIPE_FAIL_KBYTES
        cmd = ["sh", "-c", f"{send} | dd bs=1024 count={c} 2>/dev/null | sudo -n zfs receive -v -F -u -s {dst_dataset}"]
        try:
            subprocess.run(cmd, stdout=PIPE, stderr=PIPE, text=True, check=True)
        except subprocess.CalledProcessError as e:
            logging.getLogger(__name__).warning(f"generate_recv_resume_token stdout: {e.stdout}")
            self.assertIn("Partially received snapshot is saved", e.stderr)
        receive_resume_token = dataset_property(dst_dataset, "receive_resume_token")
        self.assertTrue(receive_resume_token)
        self.assertNotEqual("-", receive_resume_token)
        self.assertNotIn(dst_dataset + to_snapshot[to_snapshot.index("@") :], snapshots(dst_dataset))

    def assert_receive_resume_token(self, dataset: str, exists: bool) -> str:
        receive_resume_token = dataset_property(dataset, "receive_resume_token")
        if exists:
            self.assertTrue(receive_resume_token)
            self.assertNotEqual("-", receive_resume_token)
        else:
            self.assertEqual("-", receive_resume_token)
        return receive_resume_token


def create_filesystems(path: str, props: list[str] | None = None) -> str:
    create_filesystem(SRC_ROOT_DATASET, path, props=props)
    return create_filesystem(DST_ROOT_DATASET, path, props=props)


def recreate_filesystem(dataset: str, props: list[str] | None = None) -> str:
    if dataset_exists(dataset):
        destroy(dataset, recursive=True)
    return create_filesystem(dataset, props=props)


def create_volumes(path: str, props: list[str] | None = None) -> str:
    create_volume(SRC_ROOT_DATASET, path, size="1M", props=props)
    return create_volume(DST_ROOT_DATASET, path, size="1M", props=props)


def detect_zpool_features(location: str, pool: str) -> None:
    assert ZPOOL_FEATURES is not None
    cmd = "zpool get -Hp -o property,value all".split(" ") + [pool]
    lines = run_cmd(cmd)
    props = {line.split("\t", 1)[0]: line.split("\t", 1)[1] for line in lines}
    features = {k: v for k, v in props.items() if k.startswith("feature@")}
    ZPOOL_FEATURES[location] = features


def is_zpool_feature_enabled_or_active(location: str, feature: str) -> bool:
    assert ZPOOL_FEATURES is not None
    return ZPOOL_FEATURES[location].get(feature, "") in ("active", "enabled")


def are_bookmarks_enabled(location: str) -> bool:
    return is_zpool_feature_enabled_or_active(location, "feature@bookmark_v2") and is_zpool_feature_enabled_or_active(
        location, "feature@bookmark_written"
    )


def is_zpool_recv_resume_feature_enabled_or_active() -> bool:
    return is_zpool_feature_enabled_or_active("src", "feature@extensible_dataset") and is_zfs_at_least_2_1_0()


def is_cache_snapshots_enabled() -> bool:
    return is_zfs_at_least_2_2_0()


def fix(s: str) -> str:
    """Generate names containing leading and trailing whitespace, forbidden characters, etc."""
    return AFIX + s + AFIX


def natsorted(
    iterable: Iterable[str],
    key: Callable[[str], str] | None = None,
    reverse: bool = False,
) -> list[str]:
    """Returns a new list containing all items from the iterable in ascending order.

    If `key` is specified, it will be used to extract a comparison key from each list element.
    """
    if key is None:
        return sorted(iterable, key=natsort_key, reverse=reverse)
    else:
        return sorted(iterable, key=lambda x: natsort_key(key(x)), reverse=reverse)


def natsort_key(s: str) -> tuple[str, int, str]:
    """Sorts strings that may contain non-negative integers according to numerical value if any two strings have the same
    non-numeric prefix.

    Example: s1 < s3 < s10 < s10a < s10b
    """
    match = re.fullmatch(r"(\D*)(\d*)(.*)", s)
    if match:
        prefix, num, suffix = match.groups()
        num = int(num) if num else 0
        return prefix, num, suffix
    return s, 0, ""


def is_zfs_at_least_2_3_0() -> bool:
    ver = zfs_version()
    if ver is None:
        return False
    return is_version_at_least(ver, "2.3.0")


def is_zfs_at_least_2_2_0() -> bool:
    ver = zfs_version()
    if ver is None:
        return False
    return is_version_at_least(ver, "2.2.0")


def is_zfs_at_least_2_1_0() -> bool:
    ver = zfs_version()
    if ver is None:
        return False
    return is_version_at_least(ver, "2.1.0")


def is_pv_at_least_1_9_0() -> bool:
    ver = pv_version()
    return is_version_at_least(ver, "1.9.0")


def pv_version() -> str:
    """Example pv 1.8.5 -> 1.8.5."""
    lines = subprocess.run(["pv", "--version"], capture_output=True, text=True, check=True).stdout
    assert lines
    assert lines.startswith("pv")
    line = lines.splitlines()[0]
    version = line.split(" ")[1].strip()
    match = re.fullmatch(r"(\d+\.\d+\.\d+).*", version)
    if match:
        return match.group(1)
    else:
        raise ValueError("Unparsable pv version string: " + version)


def os_username() -> str:
    # return getpass.getuser()
    return pwd.getpwuid(os.getuid()).pw_name



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for SSH connection management utilities."""

from __future__ import (
    annotations,
)
import contextlib
import itertools
import logging
import os
import platform
import random
import shlex
import socket
import subprocess
import sys
import tempfile
import threading
import time
import unittest
from collections import (
    Counter,
)
from collections.abc import (
    Iterator,
)
from types import (
    SimpleNamespace,
)
from typing import (
    Any,
    cast,
)
from unittest.mock import (
    MagicMock,
    patch,
)

from bzfs_main import (
    bzfs,
)
from bzfs_main.configuration import (
    Params,
    Remote,
)
from bzfs_main.util import (
    connection,
)
from bzfs_main.util.connection import (
    DEDICATED,
    SHARED,
    Connection,
    ConnectionPool,
    ConnectionPools,
    dquote,
    squote,
)
from bzfs_main.util.retry import (
    RetryableError,
)
from bzfs_main.util.utils import (
    LOG_TRACE,
    sha256_urlsafe_base64,
)
from bzfs_tests.abstract_testcase import (
    AbstractTestCase,
)
from bzfs_tests.tools import (
    suppress_output,
)


#############################################################################
def suite() -> unittest.TestSuite:
    test_cases = [
        TestSimpleMiniRemote,
        TestSimpleMiniJob,
        TestQuoting,
        TestConnectionPool,
        TestRunSshCommand,
        TestTrySshCommand,
        TestRefreshSshConnection,
        TestTimeout,
        TestMaybeInjectError,
        TestDecrementInjectionCounter,
        TestSshExitOnShutdown,
    ]
    return unittest.TestSuite(unittest.TestLoader().loadTestsFromTestCase(test_case) for test_case in test_cases)


# constants:
TEST_DIR_PREFIX: str = "t_bzfs"


#############################################################################
class SlowButCorrectConnectionPool(ConnectionPool):  # validate a better implementation against this baseline

    def __init__(self, remote: Remote, max_concurrent_ssh_sessions_per_tcp_connection: int) -> None:
        super().__init__(remote, SHARED, max_concurrent_ssh_sessions_per_tcp_connection)
        self._priority_queue: list[Connection] = []  # type: ignore

    def get_connection(self) -> Connection:
        with self._lock:
            self._priority_queue.sort()
            conn = self._priority_queue[-1] if self._priority_queue else None
            if conn is None or conn._is_full():
                conn = Connection(self._remote, self._capacity)
                self._last_modified += 1
                conn._update_last_modified(self._last_modified)  # LIFO tiebreaker favors latest conn as that's most alive
                self._priority_queue.append(conn)
            conn._increment_free(-1)
            return conn

    def return_connection(self, old_conn: Connection) -> None:
        assert old_conn is not None
        with self._lock:
            assert any(old_conn is c for c in self._priority_queue)
            old_conn._increment_free(1)
            self._last_modified += 1
            old_conn._update_last_modified(self._last_modified)

    def __repr__(self) -> str:
        with self._lock:
            return str({"capacity": self._capacity, "queue_len": len(self._priority_queue), "queue": self._priority_queue})


#############################################################################
class TestConnectionPool(AbstractTestCase):

    def setUp(self) -> None:
        args = self.argparser_parse_args(args=["src", "dst", "-v", "--ssh-exit-on-shutdown"])
        p = self.make_params(args=args)
        self.src = p.src
        self.dst = p.dst
        self.dst.ssh_user_host = "127.0.0.1"
        self.remote = Remote("src", args, p)
        self.src2 = Remote("src", args, p)

    def assert_priority_queue(self, cpool: ConnectionPool, queuelen: int) -> None:
        self.assertEqual(queuelen, len(cpool._priority_queue))

    def assert_equal_connections(self, conn: Connection, donn: Connection) -> None:
        self.assertEqual(donn._ssh_cmd, conn._ssh_cmd)
        self.assertEqual(donn._free, conn._free)
        self.assertEqual(donn._last_modified, conn._last_modified)

    def get_connection(self, cpool: ConnectionPool, dpool: SlowButCorrectConnectionPool) -> tuple[Connection, Connection]:
        conn = cpool.get_connection()
        donn = dpool.get_connection()
        self.assert_equal_connections(conn, donn)
        return conn, donn

    def return_connection(
        self,
        cpool: ConnectionPool,
        conn: Connection,
        dpool: SlowButCorrectConnectionPool,
        donn: Connection,
    ) -> None:
        self.assertEqual(donn._ssh_cmd, conn._ssh_cmd)
        cpool.return_connection(conn)
        dpool.return_connection(donn)

    def test_basic(self) -> None:
        counter1a = itertools.count()
        counter2a = itertools.count()
        counter1b = itertools.count()
        self.src.local_ssh_command = lambda _socket_path=None, counter=counter1a: ([str(next(counter))], None)  # type: ignore
        self.src2.local_ssh_command = lambda _socket_path=None, counter=counter1b: ([str(next(counter))], None)  # type: ignore

        with self.assertRaises(AssertionError):
            ConnectionPool(self.src, SHARED, 0)

        capacity = 2
        cpool = ConnectionPool(self.src, SHARED, capacity)
        dpool = SlowButCorrectConnectionPool(self.src2, capacity)
        self.assert_priority_queue(cpool, 0)
        self.assertIsNotNone(repr(cpool))
        self.assertIsNotNone(str(cpool))
        cpool.shutdown("foo")

        conn1, donn1 = self.get_connection(cpool, dpool)
        self.assert_priority_queue(cpool, 1)
        self.assertEqual((capacity - 1) * 1, conn1._free)
        i = [str(next(counter2a))]
        self.assertEqual(i, conn1._ssh_cmd)
        self.assertEqual(i, conn1._ssh_cmd_quoted)
        self.assertIsNotNone(repr(conn1))
        self.assertIsNotNone(str(conn1))

        self.return_connection(cpool, conn1, dpool, donn1)
        self.assert_priority_queue(cpool, 1)

        conn2, donn2 = self.get_connection(cpool, dpool)
        self.assert_priority_queue(cpool, 1)
        self.assertIs(conn1, conn2)
        self.assertEqual((capacity - 1) * 1, conn2._free)

        conn3, donn3 = self.get_connection(cpool, dpool)
        self.assert_priority_queue(cpool, 1)
        self.assertIs(conn2, conn3)

        self.return_connection(cpool, conn2, dpool, donn2)
        self.assert_priority_queue(cpool, 1)

        self.return_connection(cpool, conn3, dpool, donn3)
        self.assert_priority_queue(cpool, 1)

        with self.assertRaises(AssertionError):
            cpool.return_connection(cast(Any, None))
        with self.assertRaises(AssertionError):
            cpool.return_connection(conn3)
        cpool.shutdown("bar")
        cpool.return_connection(conn3)

    def test_pool_acquires_lease_when_reuse_enabled(self) -> None:
        """When SSH reuse is enabled, the pool creates a single manager and acquires a lease for new connections."""
        with get_tmpdir() as ssh_socket_dir:
            args = self.argparser_parse_args(["src", "dst"])  # do not set --ssh-exit-on-shutdown
            p = self.make_params(args=args)
            remote = cast(
                Remote,
                _FakeRemote(
                    params=p,
                    location="dst",
                    ssh_user_host="host",
                    reuse_ssh_connection=True,
                    ssh_exit_on_shutdown=False,
                    ssh_control_persist_secs=90,
                    ssh_extra_opts=[],
                    ssh_socket_dir=ssh_socket_dir,
                ),
            )
            cpool = ConnectionPool(remote, SHARED, 1)
            try:
                conn = cpool.get_connection()
                self.assertIsNotNone(cpool._lease_mgr)
                self.assertIsNotNone(conn._connection_lease)
            finally:
                conn.shutdown("test")

    def test_multiple_tcp_connections(self) -> None:
        capacity = 2
        cpool = ConnectionPool(self.remote, SHARED, capacity)

        conn1 = cpool.get_connection()
        self.assertEqual((capacity - 1) * 1, conn1._free)
        conn2 = cpool.get_connection()
        self.assertIs(conn1, conn2)
        self.assertEqual((capacity - 2) * 1, conn2._free)
        self.assert_priority_queue(cpool, 1)

        conn3 = cpool.get_connection()
        self.assertIsNot(conn2, conn3)
        self.assertEqual((capacity - 1) * 1, conn3._free)
        self.assertEqual((capacity - 2) * 1, conn2._free)
        self.assert_priority_queue(cpool, 2)
        conn4 = cpool.get_connection()
        self.assertIs(conn3, conn4)
        self.assertEqual((capacity - 2) * 1, conn4._free)
        self.assert_priority_queue(cpool, 2)

        conn5 = cpool.get_connection()
        self.assertIsNot(conn4, conn5)
        self.assertEqual((capacity - 1) * 1, conn5._free)
        self.assertEqual((capacity - 2) * 1, conn4._free)
        self.assert_priority_queue(cpool, 3)

        cpool.return_connection(conn3)
        self.assert_priority_queue(cpool, 3)
        cpool.return_connection(conn4)
        t4 = cpool._last_modified
        self.assert_priority_queue(cpool, 3)

        cpool.return_connection(conn2)
        self.assert_priority_queue(cpool, 3)
        cpool.return_connection(conn1)
        t2 = cpool._last_modified
        self.assert_priority_queue(cpool, 3)

        cpool.return_connection(conn5)
        t5 = cpool._last_modified
        self.assert_priority_queue(cpool, 3)

        # assert sort order evens out the number of concurrent sessions among the TCP connections
        conn5a = cpool.get_connection()
        self.assertEqual((capacity - 1) * 1, conn5a._free)

        conn2a = cpool.get_connection()
        self.assertEqual((capacity - 1) * 1, conn2a._free)

        conn4a = cpool.get_connection()
        self.assertEqual((capacity - 1) * 1, conn4a._free)

        # assert tie-breaker in favor of most recently returned TCP connection
        conn6 = cpool.get_connection()
        self.assertEqual((capacity - 2) * 1, conn6._free)
        self.assertEqual(t5, abs(conn6._last_modified))

        conn1a = cpool.get_connection()
        self.assertEqual((capacity - 2) * 1, conn1a._free)
        self.assertEqual(t2, abs(conn1a._last_modified))

        conn4a = cpool.get_connection()
        self.assertEqual((capacity - 2) * 1, conn4a._free)
        self.assertEqual(t4, abs(conn4a._last_modified))

        cpool.shutdown("foo")

    def test_pools(self) -> None:
        with self.assertRaises(AssertionError):
            ConnectionPools(remote=self.dst, capacities={SHARED: 0})

        ConnectionPools(remote=self.dst, capacities={}).shutdown("foo")

        pools = ConnectionPools(remote=self.dst, capacities={SHARED: 8, DEDICATED: 1})
        self.assertIsNotNone(pools.pool(SHARED))
        self.assertIsNotNone(repr(pools))
        self.assertIsNotNone(str(pools))
        pools.shutdown("foo")
        pools.pool(SHARED).get_connection()
        pools.shutdown("foo")

    def test_return_sequence(self) -> None:
        maxsessions = 10
        items = 10
        for j in range(3):
            cpool = ConnectionPool(self.src, SHARED, maxsessions)
            dpool = SlowButCorrectConnectionPool(self.src2, maxsessions)
            rng = random.Random(12345)
            conns = [self.get_connection(cpool, dpool) for _ in range(items)]
            while conns:
                i = rng.randint(0, len(conns) - 1) if j == 0 else 0 if j == 1 else len(conns) - 1
                conn, donn = conns.pop(i)
                self.return_connection(cpool, conn, dpool, donn)

    def test_long_random_walk(self) -> None:
        log = logging.getLogger(bzfs.__name__)
        # loglevel = logging.DEBUG
        loglevel = LOG_TRACE
        is_logging = log.isEnabledFor(loglevel)
        num_steps = (
            25 if self.is_unit_test else 75 if self.is_smoke_test or self.is_functional_test or self.is_adhoc_test else 1000
        )
        log.info(f"num_random_steps: {num_steps}")
        start_time_nanos = time.time_ns()
        for maxsessions in range(1, 10 + 1):
            for items in range(64 + 1):
                counter1a = itertools.count()
                counter1b = itertools.count()
                self.src.local_ssh_command = lambda _socket_path=None, counter=counter1a: ([str(next(counter))], None)  # type: ignore
                self.src2.local_ssh_command = lambda _socket_path=None, counter=counter1b: ([str(next(counter))], None)  # type: ignore
                cpool = ConnectionPool(self.src, SHARED, maxsessions)
                dpool = SlowButCorrectConnectionPool(self.src2, maxsessions)
                # dpool = ConnectionPool(self.src2, SHARED, maxsessions)
                rng = random.Random(12345)
                conns = []
                try:
                    for _ in range(items):
                        conns.append(self.get_connection(cpool, dpool))
                    item = -1
                    for step in range(num_steps):
                        if is_logging:
                            log.log(loglevel, f"itr maxsessions: {maxsessions}, items: {items}, step: {step}")
                            log.log(loglevel, f"clen: {len(cpool._priority_queue)}, cpool: {cpool._priority_queue}")
                            log.log(loglevel, f"dlen: {len(dpool._priority_queue)}, dpool: {dpool._priority_queue}")
                        if not conns or rng.randint(0, 1):
                            if is_logging:
                                log.log(loglevel, "get")
                            conns.append(self.get_connection(cpool, dpool))
                        else:
                            # k = rng.randint(0, 2)
                            k = 0
                            if k == 0:
                                i = rng.randint(0, len(conns) - 1)
                            elif k == 1:
                                i = 0
                            else:
                                i = len(conns) - 1
                            conn, donn = conns.pop(i)
                            if is_logging:
                                log.log(loglevel, f"return {i}/{len(conns)+1}: conn: {conn}, donn: {donn} ")
                            self.return_connection(cpool, conn, dpool, donn)
                except Exception:
                    logging.getLogger(__name__).warning("Ooops!")
                    logging.getLogger(__name__).warning(
                        f"maxsessions: {maxsessions}, items: {items}, step: {step}, item: {item}"
                    )
                    logging.getLogger(__name__).warning(
                        f"clen: {len(cpool._priority_queue)}, cpool: {cpool._priority_queue}"
                    )
                    logging.getLogger(__name__).warning(
                        f"dlen: {len(dpool._priority_queue)}, dpool: {dpool._priority_queue}"
                    )
                    raise
                log.log(loglevel, "cpool: %s", cpool)
                # log.log(bzfs.log_debug, "cpool: %s", cpool)
        elapsed_secs = (time.time_ns() - start_time_nanos) / 1000_000_000
        log.info("random_walk took %s secs", elapsed_secs)


#############################################################################
class TestSimpleMiniRemote(AbstractTestCase):
    """Covers all control-flow branches in SimpleMiniRemote."""

    def setUp(self) -> None:
        self.log = MagicMock(logging.Logger)

    def test_init_builds_expected_options_defaults(self) -> None:
        r = connection.create_simple_miniremote(log=self.log, ssh_user_host="")
        self.assertEqual(
            ["-oBatchMode=yes", "-oServerAliveInterval=0", "-x", "-T", "-c", "^aes256-gcm@openssh.com"],
            list(r.ssh_extra_opts),
        )
        # SimpleMiniRemote always disables immediate master exit on shutdown
        self.assertFalse(r.ssh_exit_on_shutdown)
        self.assertEqual("-", r.cache_namespace())  # local mode
        self.assertEqual(([], None), r.local_ssh_command(socket_file=None))

    def test_init_with_all_extras_and_port(self) -> None:
        with get_tmpdir() as tmpdir:
            with tempfile.NamedTemporaryFile(dir=tmpdir, delete=False) as cfg:
                cfg_path = cfg.name
            r = connection.create_simple_miniremote(
                log=self.log,
                ssh_user_host="user@host",
                ssh_port=2222,
                ssh_verbose=True,
                ssh_config_file=cfg_path,
                ssh_cipher="aes128-gcm@openssh.com",
                ssh_connect_timeout_secs=60,
                reuse_ssh_connection=True,
            )
            # Verify options order and presence
            expected_prefix = [
                "-oBatchMode=yes",
                "-oServerAliveInterval=0",
                "-x",
                "-T",
                "-v",
                "-F",
                cfg_path,
                "-c",
                "aes128-gcm@openssh.com",
                "-p",
                "2222",
                "-oConnectTimeout=60s",
            ]
            self.assertEqual(expected_prefix, list(r.ssh_extra_opts))
            # cache namespace includes user@host, port and cfg hash
            expected_hash = sha256_urlsafe_base64(os.path.abspath(cfg_path), padding=False)
            self.assertEqual(f"user@host#2222#{expected_hash}", r.cache_namespace())
            # local_ssh_command includes -S only when socket provided
            cmd_no_sock, socket_path = r.local_ssh_command(socket_file=None)
            self.assertEqual([r.params.ssh_program] + expected_prefix + ["user@host"], cmd_no_sock)
            self.assertIsNone(socket_path)
            cmd_with_sock, socket_path = r.local_ssh_command(socket_file="/tmp/sock")
            self.assertEqual(
                [r.params.ssh_program] + expected_prefix + ["-S", "/tmp/sock", "user@host"],
                cmd_with_sock,
            )
            self.assertEqual("/tmp/sock", socket_path)

    def test_init_with_custom_extra_opts_is_copied(self) -> None:
        custom = ["-K"]
        r = connection.create_simple_miniremote(
            log=self.log,
            ssh_user_host="user@h",
            ssh_extra_opts=custom,
            ssh_verbose=True,
            ssh_cipher="^aes256-gcm@openssh.com",
        )
        # The instance copies and extends the provided list, leaving the caller's list untouched.
        self.assertIsNot(custom, r.ssh_extra_opts)
        self.assertEqual(["-K", "-v", "-c", "^aes256-gcm@openssh.com"], list(r.ssh_extra_opts))
        self.assertListEqual(["-K"], custom)

    def test_no_cipher_no_config_no_port(self) -> None:
        r = connection.create_simple_miniremote(
            log=self.log,
            ssh_user_host="user@h",
            ssh_cipher="",  # do not include -c
            ssh_config_file="",  # do not include -F
            ssh_port=None,  # do not include -p
        )
        opts = r.ssh_extra_opts
        self.assertNotIn("-c", opts)
        self.assertNotIn("-F", opts)
        self.assertNotIn("-p", opts)
        self.assertEqual("user@h##", r.cache_namespace())

    def test_invalid_parameters_raise_valueerror(self) -> None:
        # ssh_control_persist_secs must be >= 1
        with self.assertRaises(ValueError):
            connection.create_simple_miniremote(log=self.log, ssh_control_persist_secs=0)
        # location must be 'src' or 'dst'
        with self.assertRaises(ValueError):
            connection.create_simple_miniremote(log=self.log, location="invalid")
        with self.assertRaises(ValueError):
            connection.create_simple_miniremote(log=self.log, ssh_user_host="-oProxyCommand=evil")

    def test_invalid_ssh_program_raises_valueerror(self) -> None:
        with self.assertRaises(ValueError):
            connection.create_simple_miniremote(log=self.log, ssh_user_host="user@h", ssh_program="")

    def test_none_log_raises_valueerror(self) -> None:
        with self.assertRaises(ValueError):
            connection.create_simple_miniremote(log=cast(logging.Logger, None), ssh_user_host="user@h")

    def test_local_ssh_command_branching(self) -> None:
        r = connection.create_simple_miniremote(log=self.log, ssh_user_host="user@h", reuse_ssh_connection=False)
        # No -S because reuse is False
        self.assertEqual(
            ([r.params.ssh_program] + list(r.ssh_extra_opts) + ["user@h"], None),
            r.local_ssh_command("/tmp/s"),
        )

    def test_factory_method_round_trips_arguments(self) -> None:
        r = connection.create_simple_miniremote(
            log=self.log,
            ssh_user_host="user@host",
            ssh_port=2222,
            ssh_verbose=True,
        )
        self.assertEqual("user@host", r.ssh_user_host)
        self.assertEqual(2222, cast(Any, r).ssh_port)

    def test_simple_mini_remote_is_ssh_available_always_true(self) -> None:
        """Ensure SimpleMiniRemote.is_ssh_available always reports True."""
        r = connection.create_simple_miniremote(log=self.log, ssh_user_host="user@host")
        self.assertTrue(r.is_ssh_available())


#############################################################################
class TestSimpleMiniJob(AbstractTestCase):
    """Smoke tests for SimpleMiniJob wiring to ensure attributes are set correctly."""

    def test_initialization_defaults(self) -> None:
        j = connection.create_simple_minijob()
        self.assertIsNone(j.timeout_nanos)
        self.assertIsNone(j.timeout_duration_nanos)
        self.assertIsNotNone(j.subprocesses)

    @patch("bzfs_main.util.connection.time.monotonic_ns", return_value=1_000_000_000)
    def test_timeout_fields_from_duration(self, mock_monotonic: MagicMock) -> None:
        j = connection.create_simple_minijob(timeout_duration_secs=0.987654321)
        self.assertEqual(987654321, j.timeout_duration_nanos)
        self.assertEqual(1_000_000_000 + 987_654_321, j.timeout_nanos)
        mock_monotonic.assert_called_once_with()


#############################################################################
class TestQuoting(AbstractTestCase):
    """Covers quoting helpers used across modules."""

    def test_squote_local_no_quote(self) -> None:
        remote = MagicMock(ssh_user_host="")
        self.assertEqual("foo bar", squote(remote, "foo bar"))

    def test_squote_remote_quote(self) -> None:
        remote = MagicMock(ssh_user_host="host")
        self.assertEqual("'foo bar'", squote(remote, "foo bar"))

    def test_dquote_empty_string(self) -> None:
        """Empty arguments are quoted without escaping."""
        self.assertEqual('""', dquote(""))

    def test_dquote_preserves_single_quotes_and_spaces(self) -> None:
        """Single quotes and spaces remain untouched inside the quotes."""
        self.assertEqual("\"foo 'bar baz' qux\"", dquote("foo 'bar baz' qux"))

    def test_dquote_multiple_specials(self) -> None:
        """Every occurrence of special characters is escaped."""
        cases = [
            ('""', '"' + '\\"' * 2 + '"'),
            ("$$$", '"' + "\\$" * 3 + '"'),
            ("``", '"' + "\\`" * 2 + '"'),
        ]
        for arg, expected in cases:
            with self.subTest(arg=arg):
                self.assertEqual(expected, dquote(arg))

    def test_dquote_prevents_command_substitution(self) -> None:
        """Escaping ensures the shell prints literals rather than executing."""
        for arg in ["$(echo hacked)", "`echo hacked`"]:
            with self.subTest(arg=arg):
                self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_roundtrip_empty_string(self) -> None:
        """Ensures empty arguments survive shell evaluation unchanged."""
        arg = ""
        self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_roundtrip_with_spaces(self) -> None:
        """Verifies that spaces are preserved when passed through a shell."""
        arg = "foo bar baz"
        self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_multiple_instances(self) -> None:
        """All repeated special chars must be escaped and round-trip."""
        arg = 'aa""bb$$cc``dd'
        quoted = dquote(arg)
        self.assertEqual(2, quoted.count('\\"'))
        self.assertEqual(2, quoted.count("\\$"))
        self.assertEqual(2, quoted.count("\\`"))
        self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_preserves_unrelated_chars(self) -> None:
        """Characters outside the escape set remain intact after quoting."""
        arg = "path/with 'single' and \\backslash"
        self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_preserves_dollars_backticks_quotes(self) -> None:
        samples = [
            r"a$b",  # $ should not expand
            r"a`uname`b",  # backticks should not execute
            r'a"b',  # embedded double quote
            r"plain",  # trivial
        ]
        for s in samples:
            with self.subTest(s=s):
                self.assertEqual(s, self._sh_roundtrip(s))

    def test_dquote_handles_spaces_and_newlines(self) -> None:
        samples = [
            "a b c",  # spaces must remain a single arg
            "line1\nline2",  # newline stays literal
            " tab\tsep ",  # tabs preserved
        ]
        for s in samples:
            with self.subTest(s=s):
                self.assertEqual(s, self._sh_roundtrip(s))

    def test_dquote_does_not_require_backslash_escaping(self) -> None:
        samples = [
            r"a\b",  # normal backslash
        ]
        for s in samples:
            with self.subTest(s=s):
                self.assertEqual(s, self._sh_roundtrip(s))

    def test_dquote_globally_escapes_specials(self) -> None:
        s = "pre 'a\"b $USER `uname` \\ tail' post"
        out = dquote(s)
        payload = out[1:-1]  # strip outer double quotes

        # Every occurrence of ", $, ` in the payload must be backslash-escaped
        specials = {'"', "$", "`"}
        for i, ch in enumerate(payload):
            with self.subTest(i=i):
                if ch in specials:
                    self.assertTrue(i > 0)
                    self.assertEqual("\\", payload[i - 1], f"unescaped {ch!r} at index {i}")

    @staticmethod
    def _sh_roundtrip(arg: str) -> str:
        quoted = dquote(arg)
        return subprocess.run(["sh", "-c", f"printf %s {quoted}"], stdout=subprocess.PIPE, text=True, check=True).stdout

    @staticmethod
    def _nested_sh_roundtrip(arg: str) -> str:
        """Simulates nested 'sh -c' usage for dquote as in replication pipelines.

        Builds a command list with ``arg`` as the final argv element, joins it via ``shlex.join``, wraps it with ``dquote``
        and executes it through an outer ``sh -c`` that invokes an inner ``sh -c``. Returns the string observed by the inner
        Python process so tests can assert end-to-end preservation.
        """
        recv_cmd: list[str] = [sys.executable, "-c", "import sys; print(sys.argv[-1])", arg]
        recv_cmd_str: str = shlex.join(recv_cmd)
        script: str = dquote(recv_cmd_str)
        stdout: str = subprocess.run(["sh", "-c", f"sh -c {script}"], stdout=subprocess.PIPE, text=True, check=True).stdout
        return stdout.rstrip("\n")

    def test_dquote_backslash_before_quote(self) -> None:
        """Backslashes immediately before a quote survive shell evaluation."""
        arg = r"a\"b"
        self.assertEqual(arg, self._sh_roundtrip(arg))

    def test_dquote_escapes(self) -> None:
        self.assertEqual('"a\\"b\\$c\\`d"', dquote('a"b$c`d'))

    def test_dquote_nested_sh_backslash_backtick(self) -> None:
        """Nested sh -c round-trips values with backslash+backtick."""
        value = r"/tmp/foo\`bar"
        prop_arg = f"prop={value}"
        self.assertEqual(prop_arg, self._nested_sh_roundtrip(prop_arg))

    def test_dquote_preserves_double_backslashes_in_nested_sh(self) -> None:
        """Ensures nested sh -c round-trips values containing double backslashes."""
        samples = [
            r"\\server\share",  # UNC-style path
            r"\\path\\with\\many",  # multiple segments
        ]
        for value in samples:
            with self.subTest(value=value):
                prop_arg: str = f"foo={value}"
                self.assertEqual(prop_arg, self._nested_sh_roundtrip(prop_arg))


def make_fake_params() -> Params:
    mock = MagicMock(spec=Params)
    mock.log = logging.getLogger(__name__)
    mock.ssh_program = "ssh"
    mock.connection_pools = {}
    return mock


#############################################################################
class _FakeRemote(SimpleNamespace):
    ssh_extra_opts: tuple[str, ...]

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not hasattr(self, "params"):
            self.params = make_fake_params()
        if not hasattr(self, "ssh_extra_opts"):
            self.ssh_extra_opts = ()
        if not hasattr(self, "ssh_socket_dir"):
            self.ssh_socket_dir = tempfile.mkdtemp(prefix=TEST_DIR_PREFIX)
        if not hasattr(self, "ssh_control_persist_margin_secs"):
            self.ssh_control_persist_margin_secs = 2
        if not hasattr(self, "cache_namespace"):
            self.cache_namespace = lambda: "user@host#22#cfg"
        if not hasattr(self, "pool"):
            self.pool = SHARED

    def local_ssh_command(self, socket_path: str | None = None) -> tuple[list[str], str | None]:
        return ["ssh", self.ssh_user_host or "localhost"], socket_path

    def is_ssh_available(self) -> bool:
        return True


#############################################################################
class TestRunSshCommand(AbstractTestCase):

    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.params = make_fake_params()
        self.job.timeout_nanos = None
        self.remote = cast(
            Remote,
            _FakeRemote(
                location="dst", ssh_user_host="", reuse_ssh_connection=True, ssh_control_persist_secs=90, ssh_extra_opts=[]
            ),
        )
        self.conn_pools = ConnectionPools(remote=self.remote, capacities={SHARED: 1})
        self.conn_pool = self.conn_pools.pool(SHARED)
        self.conn: Connection | None = None

        original_connection = self.conn_pool.connection

        @contextlib.contextmanager
        def _connection_cm(original_connection: Any = original_connection) -> Iterator[Connection]:
            with original_connection() as conn:
                self.conn = conn
                yield conn

        # Wrap connection() so tests can assert it was used.
        self.conn_pool_connection_mock = MagicMock(side_effect=_connection_cm)
        self.conn_pool.connection = self.conn_pool_connection_mock  # type: ignore[method-assign]  # replace context manager with mock
        self.job.params.connection_pools = {"dst": self.conn_pools}

    def test_empty_cmd_raises_value_error(self) -> None:
        job = connection.create_simple_minijob()
        conn = Connection(self.remote, 1)
        with self.assertRaises(ValueError):
            conn.run_ssh_command(cmd=[], job=job)

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_dry_run_skips_execution(self, mock_run: MagicMock) -> None:
        result = self.job.run_ssh_command(self.remote, cmd=["ls"], is_dry=True)
        self.assertEqual("", result)
        mock_run.assert_not_called()
        self.conn_pool_connection_mock.assert_called_once()

    @patch("bzfs_main.util.connection.Connection.refresh_ssh_connection_if_necessary", autospec=True)
    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_remote_calls_refresh_and_executes(self, mock_run: MagicMock, mock_refresh: MagicMock) -> None:
        self.remote.ssh_user_host = "host"
        self.conn_pool._remote.ssh_user_host = "host"
        mock_run.return_value = MagicMock(stdout="out", stderr="err")
        result = self.job.run_ssh_command(self.remote, cmd=["echo"], print_stdout=True, print_stderr=True)
        self.assertEqual("out", result)
        assert self.conn is not None
        mock_refresh.assert_called_once_with(self.conn, self.job)
        mock_run.assert_called_once()
        self.conn_pool_connection_mock.assert_called_once()

    @patch("bzfs_main.util.connection.Connection.refresh_ssh_connection_if_necessary", autospec=True)
    @patch(
        "bzfs_main.util.utils.Subprocesses.subprocess_run",
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="cmd", output="o", stderr="e"),
    )
    def test_calledprocesserror_propagates(self, mock_run: MagicMock, mock_refresh: MagicMock) -> None:
        self.remote.ssh_user_host = "h"
        self.conn_pool._remote.ssh_user_host = "h"
        with self.assertRaises(subprocess.CalledProcessError):
            self.job.run_ssh_command(self.remote, cmd=["boom"], print_stdout=True)
        mock_run.assert_called_once()
        self.conn_pool_connection_mock.assert_called_once()

    def test_real_ssh_transport_error_raises_retryable_error(self) -> None:
        """Uses the real ssh CLI to trigger a transport error and exercise RetryableError wrapping."""
        # Use real Subprocesses.subprocess_run (no patch) and a host name that cannot resolve.
        self.remote.ssh_user_host = ""
        # Ensure the underlying Connection uses a short timeout for the failing ssh call.
        with self.conn_pool.connection() as conn:
            conn._ssh_cmd = ["ssh", "-T", "-x", "-oBatchMode=yes", "-oConnectTimeout=1"]  # type: ignore[misc]  # override test-only ssh_cmd
        self.conn_pool_connection_mock.reset_mock()

        with self.assertRaises(RetryableError) as cm, suppress_output():
            # Pass an invalid host name as the first argument so ssh treats it as the remote host.
            self.job.run_ssh_command(self.remote, cmd=["nonexistent.invalid"], print_stdout=False, print_stderr=True)

        self.conn_pool_connection_mock.assert_called_once()
        self.assertEqual("ssh", cm.exception.display_msg)
        cause = cm.exception.__cause__
        self.assertIsNotNone(cause)
        assert isinstance(cause, subprocess.CalledProcessError)
        self.assertEqual(255, cause.returncode)

    @patch("bzfs_main.util.connection.Connection.refresh_ssh_connection_if_necessary", autospec=True)
    @patch(
        "bzfs_main.util.utils.Subprocesses.subprocess_run",
        side_effect=subprocess.TimeoutExpired(cmd="cmd", timeout=1),
    )
    def test_timeout_expired_propagates(self, mock_run: MagicMock, mock_refresh: MagicMock) -> None:
        self.remote.ssh_user_host = "h"
        self.conn_pool._remote.ssh_user_host = "h"
        with self.assertRaises(subprocess.TimeoutExpired):
            self.job.run_ssh_command(self.remote, cmd=["sleep"], print_stdout=True)
        mock_run.assert_called_once()
        self.conn_pool_connection_mock.assert_called_once()

    @patch("bzfs_main.util.connection.Connection.refresh_ssh_connection_if_necessary", autospec=True)
    @patch(
        "bzfs_main.util.utils.Subprocesses.subprocess_run",
        side_effect=UnicodeDecodeError("utf-8", b"x", 0, 1, "boom"),
    )
    def test_unicode_error_propagates_without_logging(self, mock_run: MagicMock, mock_refresh: MagicMock) -> None:
        self.remote.ssh_user_host = "h"
        self.conn_pool._remote.ssh_user_host = "h"
        with self.assertRaises(UnicodeDecodeError):
            self.job.run_ssh_command(self.remote, cmd=["foo"])
        mock_run.assert_called_once()
        self.conn_pool_connection_mock.assert_called_once()


#############################################################################
class TestTrySshCommand(AbstractTestCase):

    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.params = make_fake_params()
        self.remote = cast(
            Remote, _FakeRemote(location="dst", ssh_user_host="host", reuse_ssh_connection=True, ssh_extra_opts=[])
        )

    @patch("bzfs_main.bzfs.Job.run_ssh_command", return_value="ok")
    @patch("bzfs_main.bzfs.Job.maybe_inject_error")
    def test_success(self, mock_inject: MagicMock, mock_run: MagicMock) -> None:
        result = self.job.try_ssh_command(self.remote, loglevel=0, cmd=["ls"])
        self.assertEqual("ok", result)
        mock_inject.assert_called_once()
        mock_run.assert_called_once()

    @patch(
        "bzfs_main.bzfs.Job.run_ssh_command",
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="cmd", stderr="zfs: dataset does not exist"),
    )
    def test_dataset_missing_returns_none(self, mock_run: MagicMock) -> None:
        result = self.job.try_ssh_command(self.remote, loglevel=0, cmd=["zfs"], exists=True)
        self.assertIsNone(result)
        mock_run.assert_called_once()

    @patch(
        "bzfs_main.bzfs.Job.run_ssh_command",
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="cmd", stderr="boom"),
    )
    def test_other_error_raises_retryable(self, mock_run: MagicMock) -> None:
        with self.assertRaises(RetryableError):
            self.job.try_ssh_command(self.remote, loglevel=0, cmd=["zfs"], exists=False)
        mock_run.assert_called_once()

    @patch(
        "bzfs_main.bzfs.Job.run_ssh_command",
        side_effect=UnicodeDecodeError("utf-8", b"x", 0, 1, "boom"),
    )
    def test_unicode_error_wrapped(self, mock_run: MagicMock) -> None:
        with self.assertRaises(RetryableError):
            self.job.try_ssh_command(self.remote, loglevel=0, cmd=["x"])
        mock_run.assert_called_once()

    @patch("bzfs_main.bzfs.Job.maybe_inject_error", side_effect=RetryableError("Subprocess failed"))
    def test_injected_retryable_error_propagates(self, mock_inject: MagicMock) -> None:
        with self.assertRaises(RetryableError):
            self.job.try_ssh_command(self.remote, loglevel=0, cmd=["x"])
        mock_inject.assert_called_once()


#############################################################################
class TestRefreshSshConnection(AbstractTestCase):
    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.params = make_fake_params()
        self.job.timeout_nanos = None
        self.remote = cast(
            Remote,
            _FakeRemote(
                params=self.job.params,
                location="dst",
                ssh_user_host="host",
                reuse_ssh_connection=True,
                ssh_control_persist_secs=4,
                ssh_exit_on_shutdown=True,
                ssh_extra_opts=[],
            ),
        )
        self.remote.ssh_control_persist_margin_secs = 1
        self.conn = connection.Connection(self.remote, 1)

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_local_mode_returns_immediately(self, mock_run: MagicMock) -> None:
        self.remote.ssh_user_host = ""
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_run.assert_not_called()

    @patch("bzfs_main.util.connection.die", side_effect=SystemExit)
    def test_ssh_unavailable_dies(self, mock_die: MagicMock) -> None:
        self.remote.is_ssh_available = MagicMock(return_value=False)  # type: ignore[method-assign]
        with self.assertRaises(SystemExit):
            self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_die.assert_called_once()

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_reuse_disabled_no_action(self, mock_run: MagicMock) -> None:
        self.remote.reuse_ssh_connection = False
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_run.assert_not_called()

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_connection_alive_fast_path(self, mock_run: MagicMock) -> None:
        self.conn._last_refresh_time = time.monotonic_ns()
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_run.assert_not_called()

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_master_alive_refreshes_timestamp(self, mock_run: MagicMock) -> None:
        mock_run.return_value = MagicMock(returncode=0)
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_run.assert_called_once()
        self.assertGreater(self.conn._last_refresh_time, 0)

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_master_start_failure_is_retryable(self, mock_run: MagicMock) -> None:
        # First call: '-O check' returns non-zero, indicating master not alive.
        # Second call: starting master with check=True raises CalledProcessError.
        mock_run.side_effect = [
            MagicMock(returncode=1),
            subprocess.CalledProcessError(returncode=1, cmd="ssh", stderr="bad"),
        ]
        with self.assertRaises(RetryableError):
            self.conn.refresh_ssh_connection_if_necessary(self.job)

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_verbose_option_limits_persist_time(self, mock_run: MagicMock) -> None:
        self.remote.ssh_extra_opts = ("-v",)
        mock_run.side_effect = [MagicMock(returncode=1), MagicMock(returncode=0)]
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        args = mock_run.call_args_list[1][0][0]
        self.assertIn("-oControlPersist=1s", args)

    @patch("bzfs_main.util.connection_lease.ConnectionLease.set_socket_mtime_to_now")
    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_mtime_now_invoked_when_connection_lease_present(self, mock_run: MagicMock, mock_utime: MagicMock) -> None:
        """Ensures set_socket_mtime_to_now() is called when a Connection has a lease (socket path).

        Uses a remote with ssh_exit_on_shutdown=False so Connection acquires a lease in __init__. Mocks subprocess_run to the
        fast 'alive' path (returncode=0).
        """
        mock_run.return_value = MagicMock(returncode=0)
        with get_tmpdir() as ssh_socket_dir:
            remote = cast(
                Remote,
                _FakeRemote(
                    params=self.job.params,
                    location="dst",
                    ssh_user_host="host",
                    reuse_ssh_connection=True,
                    ssh_control_persist_secs=4,
                    ssh_exit_on_shutdown=False,  # acquire lease
                    ssh_extra_opts=[],
                    ssh_socket_dir=ssh_socket_dir,
                ),
            )
            cpool = ConnectionPool(remote, SHARED, 1)
            conn = cpool.get_connection()
            try:
                conn.refresh_ssh_connection_if_necessary(self.job)
                self.assertTrue(mock_utime.called)
                self.assertIsNotNone(conn._connection_lease)
                mock_utime.assert_called_once()
            finally:
                conn.shutdown("test")

    @patch("bzfs_main.util.connection_lease.ConnectionLease.set_socket_mtime_to_now")
    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_mtime_now_not_invoked_when_no_connection_lease(self, mock_run: MagicMock, mock_utime: MagicMock) -> None:
        """Ensures set_socket_mtime_to_now() is not called when a Connection has no lease (lease is None)."""
        mock_run.return_value = MagicMock(returncode=0)
        # In setUp, ssh_exit_on_shutdown=True ensures no lease is acquired.
        self.assertIsNone(self.conn._connection_lease)
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_utime.assert_not_called()

    @patch("bzfs_main.util.connection.time.monotonic_ns", return_value=-(2**64))
    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_fresh_boot_does_not_skip_first_refresh(self, mock_run: MagicMock, _mock_time: MagicMock) -> None:
        """Always perform the first connection refresh even when monotonic clock is tiny after operating system reboot."""
        mock_run.return_value = MagicMock(returncode=0)
        self.conn.refresh_ssh_connection_if_necessary(self.job)
        mock_run.assert_called_once()
        argv = mock_run.call_args[0][0]
        self.assertIn("-O", argv)
        self.assertIn("check", argv)

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_missing_control_socket_rechecks_even_in_fast_path(self, mock_run: MagicMock) -> None:
        """Recheck immediately when the expected control socket path is missing.

        Purpose: avoid waiting almost the full ControlPersist window after a transient master loss.
        Assumption: Connection has a lease-backed socket path.
        Rationale: missing socket is a cheap local signal that the fast path should be bypassed.
        """
        mock_run.return_value = MagicMock(returncode=0)
        with get_tmpdir() as ssh_socket_dir:
            remote = connection.create_simple_miniremote(
                log=self.job.params.log,
                ssh_user_host="host",
                reuse_ssh_connection=True,
                ssh_control_persist_secs=4,
                ssh_control_persist_margin_secs=1,
                ssh_socket_dir=ssh_socket_dir,
            )
            cpool = ConnectionPool(remote, SHARED, 1)
            conn = cpool.get_connection()
            try:
                lease = conn._connection_lease
                self.assertIsNotNone(lease)
                assert lease is not None
                conn._last_refresh_time = time.monotonic_ns()  # force old fast-path condition
                self.assertFalse(os.path.exists(lease.socket_path))
                conn.refresh_ssh_connection_if_necessary(self.job)
                mock_run.assert_called_once()
                argv = mock_run.call_args[0][0]
                self.assertIn("-O", argv)
                self.assertIn("check", argv)
            finally:
                conn.shutdown("test")

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_stale_control_socket_rechecks_even_in_fast_path(self, mock_run: MagicMock) -> None:
        """Recheck immediately when the control socket path exists but is stale.

        Purpose: avoid long slow-path periods when socket inode exists without a listening master.
        Assumption: stale sockets can survive abnormal process termination.
        Rationale: socket connectability is a stronger liveness signal than mere path existence.
        """
        mock_run.return_value = MagicMock(returncode=0)
        with get_tmpdir() as ssh_socket_dir:
            remote = connection.create_simple_miniremote(
                log=self.job.params.log,
                ssh_user_host="host",
                reuse_ssh_connection=True,
                ssh_control_persist_secs=4,
                ssh_control_persist_margin_secs=1,
                ssh_socket_dir=ssh_socket_dir,
            )
            cpool = ConnectionPool(remote, SHARED, 1)
            conn = cpool.get_connection()
            try:
                lease = conn._connection_lease
                self.assertIsNotNone(lease)
                assert lease is not None
                stale_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                stale_sock.bind(lease.socket_path)
                stale_sock.close()  # leaves behind a stale socket path without a listener
                conn._last_refresh_time = time.monotonic_ns()  # force old fast-path condition
                self.assertTrue(os.path.exists(lease.socket_path))
                conn.refresh_ssh_connection_if_necessary(self.job)
                mock_run.assert_called_once()
                argv = mock_run.call_args[0][0]
                self.assertIn("-O", argv)
                self.assertIn("check", argv)
            finally:
                conn.shutdown("test")

    @patch("bzfs_main.util.utils.Subprocesses.subprocess_run")
    def test_stale_control_socket_is_removed_before_master_restart(self, mock_run: MagicMock) -> None:
        """Remove stale control socket path before starting a new SSH master.

        Purpose: prevent OpenSSH from disabling multiplexing when path exists but listener is gone.
        Assumption: stale socket path can remain after abnormal termination.
        Rationale: unlinking stale path lets `ssh -M -S` create a fresh listening control socket.
        """
        mock_run.side_effect = [MagicMock(returncode=1), MagicMock(returncode=0)]
        with get_tmpdir() as ssh_socket_dir:
            remote = connection.create_simple_miniremote(
                log=self.job.params.log,
                ssh_user_host="host",
                reuse_ssh_connection=True,
                ssh_control_persist_secs=4,
                ssh_control_persist_margin_secs=1,
                ssh_socket_dir=ssh_socket_dir,
            )
            cpool = ConnectionPool(remote, SHARED, 1)
            conn = cpool.get_connection()
            try:
                lease = conn._connection_lease
                self.assertIsNotNone(lease)
                assert lease is not None
                stale_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                stale_sock.bind(lease.socket_path)
                stale_sock.close()
                conn._last_refresh_time = time.monotonic_ns()  # force fast-window condition
                conn.refresh_ssh_connection_if_necessary(self.job)
                self.assertFalse(os.path.exists(lease.socket_path))
                self.assertEqual(2, mock_run.call_count)
                start_argv = mock_run.call_args_list[1][0][0]
                self.assertIn("-M", start_argv)
            finally:
                conn.shutdown("test")


#############################################################################
class TestTimeout(AbstractTestCase):

    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.params = make_fake_params()

    def test_no_timeout_returns_none(self) -> None:
        self.assertIsNone(connection.timeout(self.job))
        self.job.timeout_nanos = None
        self.job.timeout_duration_nanos = None
        self.assertIsNone(connection.timeout(self.job))

    def test_expired_timeout_raises(self) -> None:
        self.job.timeout_duration_nanos = 1
        self.job.timeout_nanos = time.monotonic_ns() - self.job.timeout_duration_nanos
        with self.assertRaises(subprocess.TimeoutExpired):
            connection.timeout(self.job)

    def test_seconds_remaining_returned(self) -> None:
        self.job.timeout_duration_nanos = 1_000_000_000
        self.job.timeout_nanos = time.monotonic_ns() + self.job.timeout_duration_nanos
        result = connection.timeout(self.job)
        self.assertIsNotNone(result)
        self.assertGreaterEqual(cast(float, result), 0.0)


#############################################################################
class TestMaybeInjectError(AbstractTestCase):

    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.injection_lock = threading.Lock()
        self.job.error_injection_triggers = {"before": Counter()}

    def test_no_trigger_noop(self) -> None:
        self.job.maybe_inject_error(cmd=["cmd"])

    def test_missing_counter_noop(self) -> None:
        self.job.maybe_inject_error(cmd=["cmd"], error_trigger="boom")

    def test_counter_zero_noop(self) -> None:
        self.job.error_injection_triggers["before"]["boom"] = 0
        self.job.maybe_inject_error(cmd=["cmd"], error_trigger="boom")
        self.assertEqual(0, self.job.error_injection_triggers["before"]["boom"])

    def test_nonretryable_error_raised(self) -> None:
        self.job.error_injection_triggers["before"]["boom"] = 1
        with self.assertRaises(subprocess.CalledProcessError):
            self.job.maybe_inject_error(cmd=["cmd"], error_trigger="boom")
        self.assertEqual(0, self.job.error_injection_triggers["before"]["boom"])

    def test_retryable_error_raised(self) -> None:
        self.job.error_injection_triggers["before"]["retryable_boom"] = 1
        with self.assertRaises(RetryableError):
            self.job.maybe_inject_error(cmd=["cmd"], error_trigger="retryable_boom")


#############################################################################
class TestDecrementInjectionCounter(AbstractTestCase):

    def setUp(self) -> None:
        self.job = bzfs.Job()
        self.job.injection_lock = threading.Lock()

    def test_zero_counter_returns_false(self) -> None:
        counter: Counter = Counter()
        self.assertFalse(self.job.decrement_injection_counter(counter=counter, trigger="x"))

    def test_positive_counter_decrements(self) -> None:
        counter: Counter = Counter({"x": 2})
        self.assertTrue(self.job.decrement_injection_counter(counter=counter, trigger="x"))
        self.assertEqual(1, counter["x"])


#############################################################################
class TestSshExitOnShutdown(AbstractTestCase):

    @patch("bzfs_main.util.connection.subprocess.run")
    def test_ssh_exit_on_shutdown_true_triggers_exit(self, mock_run: MagicMock) -> None:
        """Purpose: Validate that when ``--ssh-exit-on-shutdown`` is enabled, the connection shutdown triggers an SSH
        ControlMaster administrative ``-O exit`` request, asking the master to terminate immediately once idle.

        Assumptions: The test substitutes ``subprocess.run`` with a mock to avoid invoking real SSH. The remote's
        configuration has exit-on-shutdown enabled and does not attempt ControlPath reuse, so a targeted control
        command is appropriate. The CLI argument parser populates the parameter object correctly.

        Design Rationale: Inspecting the captured argv is the most reliable and fast way to validate orchestration
        logic without requiring a live SSH daemon. The test focuses on the presence of ``-O`` and ``exit`` flags rather
        than full command equivalence to remain resilient to harmless option reordering. This creates a stable,
        low-flake assertion that still proves the intended behavior is wired end-to-end.
        """

        args = self.argparser_parse_args(["src", "dst"])
        p = self.make_params(args=args)
        r = cast(
            Remote,
            _FakeRemote(
                params=p,
                location="src",
                ssh_user_host="host",
                reuse_ssh_connection=True,
                ssh_exit_on_shutdown=True,
                ssh_extra_opts=[],
            ),
        )
        conn: Connection = Connection(r, 1)
        conn.shutdown("test")
        self.assertTrue(mock_run.called)
        argv = mock_run.call_args[0][0]
        self.assertIn("-O", argv)
        self.assertIn("exit", argv)

    @patch("bzfs_main.util.connection.subprocess.run")
    def test_ssh_exit_on_shutdown_false_skips_exit(self, mock_run: MagicMock) -> None:
        """Purpose: Ensure that the default behavior, and explicit disabling of immediate exit, avoids sending the
        ``-O exit`` administrative command to an SSH ControlMaster during shutdown. Reusable masters should persist to
        amortize connection setup costs across jobs.

        Assumptions: The test runs with ``reuse_ssh_connection=True`` and without the ``--ssh-exit-on-shutdown`` flag.
        ``subprocess.run`` is mocked to prevent external effects and to make call absence observable. Socket directories
        are created in a temporary location and not shared with other tests.

        Design Rationale: Negative tests guard against accidental regressions that would prematurely tear down masters,
        harming performance and stability. By simply asserting the mock was not called, the test remains robust and
        independent of incidental command formatting or environment. This keeps the specification crisp: in this mode,
        shutdown should be a no-op with respect to master termination.
        """
        with get_tmpdir() as ssh_socket_dir:
            args = self.argparser_parse_args(["src", "dst"])  # default is to NOT exit masters immediately
            p = self.make_params(args=args)
            r = cast(
                Remote,
                _FakeRemote(
                    params=p,
                    location="src",
                    ssh_user_host="host",
                    reuse_ssh_connection=True,
                    ssh_exit_on_shutdown=False,
                    ssh_control_persist_secs=90,
                    ssh_extra_opts=[],
                    ssh_socket_dir=ssh_socket_dir,
                ),
            )
            cpool = ConnectionPool(r, SHARED, 1)
            conn: Connection = cpool.get_connection()
            conn.shutdown("test")
            # No call to run expected when immediate exit is false
            self.assertFalse(mock_run.called)


def get_tmpdir() -> tempfile.TemporaryDirectory[str]:
    # on OSX the default test tmp dir is unreasonably long, e.g.
    # /var/folders/dj/y5mjnplj0l79dq4dkknbsjbh0000gp/T/tmp7j51fpp/
    root_dir = "/tmp" if platform.system() == "Darwin" else None
    return tempfile.TemporaryDirectory(dir=root_dir)
